# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Test the `Channel` and `ChannelList` objects."""

from unittest import mock

import numpy
import pytest
from astropy import units

from ...segments import (
    Segment,
    SegmentList,
)
from ...testing import (
    mocks,
    utils,
)
from .. import (
    Channel,
    ChannelList,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

OMEGA_CONFIG = """
[L1:CAL,L1 calibrated]

{
  channelName:                 'L1:GDS-CALIB_STRAIN'
  frameType:                   'L1_HOFT_C00'
  sampleFrequency:             4096
  searchTimeRange:             64
  searchFrequencyRange:        [0 Inf]
  searchQRange:                [4 96]
  searchMaximumEnergyLoss:     0.2
  whiteNoiseFalseRate:         1e-3
  searchWindowDuration:        0.5
  plotTimeRanges:              [1 4 16]
  plotFrequencyRange:          []
  plotNormalizedEnergyRange:   [0 25.5]
  alwaysPlotFlag:              1
}

[L1:PEM,L1 environment]

{
  channelName:                 'L1:PEM-CS_SEIS_LVEA_VERTEX_Z_DQ'
  frameType:                   'L1_R'
  sampleFrequency:             128
  searchTimeRange:             1024
  searchFrequencyRange:        [0 Inf]
  searchQRange:                [4 64]
  searchMaximumEnergyLoss:     0.2
  whiteNoiseFalseRate:         1e-3
  searchWindowDuration:        1.0
  plotTimeRanges:              [8 64 512]
  plotFrequencyRange:          []
  plotNormalizedEnergyRange:   [0 25.5]
  alwaysPlotFlag:              0
}
"""

CLF = """
[group-1]
flow = 4
fhigh = Nyquist
qhigh = 150
frametype = H1_HOFT_C00
channels = H1:GDS-CALIB_STRAIN 16384 unsafe clean

[group-2]
flow = .1
fhigh = 60
qhigh = 60
frametype = H1_R
channels =
    H1:ISI-GND_STS_HAM2_X_DQ 512 safe flat
    H1:ISI-GND_STS_HAM2_Y_DQ 256 unsafe flat
    H1:ISI-GND_STS_HAM2_Z_DQ 512 glitchy
"""


# -- Channel -------------------------

class TestChannel:
    """Test `Channel`."""

    TEST_CLASS = Channel

    # -- test creation ---------------

    def test_empty(self):
        """Test `Channel` with no params."""
        new = self.TEST_CLASS("")
        assert str(new) == ""
        assert new.sample_rate is None
        assert new.dtype is None

    def test_new(self):
        """Test `Channel` creation with params."""
        new = self.TEST_CLASS(
            "X1:GWPY-TEST_CHANNEL_NAME",
            sample_rate=64,
            unit="m",
        )
        assert str(new) == "X1:GWPY-TEST_CHANNEL_NAME"
        assert new.ifo == "X1"
        assert new.system == "GWPY"
        assert new.subsystem == "TEST"
        assert new.signal == "CHANNEL_NAME"
        assert new.sample_rate == 64 * units.Hz
        assert new.unit is units.meter

        new2 = self.TEST_CLASS(new)
        assert new2.sample_rate == new.sample_rate
        assert new2.unit == new.unit
        assert new2.texname == new.texname

    # -- test properties -------------

    @pytest.mark.parametrize(("arg", "fs"), [
        pytest.param(None, None, id="None"),
        pytest.param(1, 1 * units.Hz, id="float"),
        pytest.param(1 * units.Hz, 1 * units.Hz, id="hz"),
        pytest.param(1000 * units.mHz, 1 * units.Hz, id="mHz"),
        pytest.param("1", 1 * units.Hz, id="string"),
    ])
    def test_sample_rate(self, arg, fs):
        """Test `Channel.sample_rate`."""
        new = self.TEST_CLASS("test", sample_rate=arg)
        if arg is not None:
            assert isinstance(new.sample_rate, units.Quantity)
        assert new.sample_rate == fs

    @pytest.mark.parametrize(("arg", "unit"), [
        (None, None),
        ("m", units.m),
    ])
    def test_unit(self, arg, unit):
        """Test `Channel.unit`."""
        new = self.TEST_CLASS("test", unit=arg)
        if arg is not None:
            assert isinstance(new.unit, units.UnitBase)
        assert new.unit == unit

    def test_frequency_range(self):
        """Test `Channel.frequency_range`."""
        new = self.TEST_CLASS("test", frequency_range=(1, 40))
        assert isinstance(new.frequency_range, units.Quantity)
        utils.assert_quantity_equal(new.frequency_range, (1, 40) * units.Hz)

        with pytest.raises(TypeError):
            Channel("", frequency_range=1)

    def test_safe(self):
        """Test `Channel.safe`."""
        new = self.TEST_CLASS("")
        assert new.safe is None

    @pytest.mark.parametrize(("value", "result"), [
        pytest.param(None, None, id="None"),
        pytest.param(True, True, id="True"),
        pytest.param(1, True, id="truthy"),
    ])
    def test_safe_setter(self, value, result):
        """Test `Channel.safe` property setting."""
        new = self.TEST_CLASS("")
        new.safe = value  # type: ignore[assignment]
        assert new.safe == result

    @pytest.mark.parametrize(("arg", "model"), [
        (None, None),
        ("H1ASCIMC", "h1ascimc"),
    ])
    def test_model(self, arg, model):
        """Test `Channel.model`."""
        new = self.TEST_CLASS("test", model=arg)
        assert new.model == model

    @pytest.mark.parametrize(("arg", "type_", "ndstype"), [
        pytest.param(None, None, 0, id="unknown"),
        pytest.param("m-trend", "m-trend", 16, id="m-trend"),
        pytest.param(8, "s-trend", 8, id="s-trend"),
    ])
    def test_type_ndstype(self, arg, type_, ndstype):
        """Test `Channel.ndstype`."""
        c = self.TEST_CLASS("", type=arg)
        assert c.type == type_
        assert c.ndstype == ndstype

    def test_type_ndstype_unknown(self):
        """Test `Channel.ndstype` 'unknown'."""
        assert self.TEST_CLASS("", type="blah").type == "UNKNOWN"
        assert self.TEST_CLASS("", type="blah").ndstype == 0

    @pytest.mark.parametrize(("arg", "dtype"), [
        (None, None),
        (16, numpy.dtype("float64")),
        (float, numpy.dtype("float64")),
        ("float", numpy.dtype("float64")),
        ("float64", numpy.dtype("float64")),
        ("u4", numpy.dtype("uint32")),
        (bool, numpy.dtype("bool")),
    ])
    def test_dtype(self, arg, dtype):
        """Test `Channel.dtype`."""
        new = self.TEST_CLASS("test", dtype=arg)
        assert new.dtype is dtype

    @pytest.mark.parametrize("url", [
        None,
        "https://blah",
        "file://local/path",
    ])
    def test_url(self, url):
        """Test `Channel.url`."""
        new = self.TEST_CLASS("test", url=url)
        assert new.url == url

    @pytest.mark.parametrize("url", [
        "BAD",
        "ftp://example.com/data",
    ])
    def test_url_error(self, url):
        """Test `Channel.url` error handling."""
        with pytest.raises(ValueError, match=fr"^Invalid URL {url!r}$"):
            self.TEST_CLASS("test", url=url)

    def test_frametype(self):
        """Test `Channel.frametype`."""
        new = self.TEST_CLASS("test", frametype="BLAH")
        assert new.frametype == "BLAH"

    @pytest.mark.parametrize(("name", "texname"), [
        ("X1:TEST", "X1:TEST"),
        ("X1:TEST-CHANNEL_NAME", r"X1:TEST-CHANNEL\_NAME"),
    ])
    def test_texname(self, name, texname):
        """Test `Channel.texname`."""
        new = self.TEST_CLASS(name)
        assert new.texname == texname

    @pytest.mark.parametrize(("ndstype", "ndsname"), [
        (None, "X1:TEST"),
        ("m-trend", "X1:TEST,m-trend"),
        ("raw", "X1:TEST"),
    ])
    def test_ndsname(self, ndstype, ndsname):
        """Test `Channel.ndsname`."""
        new = self.TEST_CLASS("X1:TEST", type=ndstype)
        assert new.ndsname == ndsname

    # -- test magic ------------------

    @pytest.mark.parametrize(("name", "result"), [
        ("X1:TEST", "X1:TEST"),
        ("X1:TEST,m-trend", "X1:TEST"),
    ])
    def test_str(self, name, result):
        """Test `str(channel)`."""
        assert str(self.TEST_CLASS(name)) == result

    @mock.patch("builtins.hex", lambda x: 12345)  # noqa: ARG005,PT008
    def test_repr(self):
        """Test `repr(channel)`."""
        new = self.TEST_CLASS(
            "X1:TEST-CHANNEL",
            type=8,
            sample_rate=16384,
        )
        assert repr(new) == (
            f'<{type(new).__name__}("{new}" [s-trend], 16384.0 Hz) at 12345>'
        )

    @pytest.mark.parametrize(("a", "b", "result"), [
        pytest.param(
            TEST_CLASS("test"),
            TEST_CLASS("test"),
            True,
            id="channel match",
        ),
        pytest.param(
            TEST_CLASS("test", sample_rate=2),
            TEST_CLASS("test", sample_rate=4),
            False,
            id="channel mismatch",
        ),
        pytest.param(
            TEST_CLASS("test"),
            None,
            False,
            id="wrong type",
        ),
    ])
    def test_eq(self, a, b, result):
        """Test `Channel.__eq__`."""
        assert a.__eq__(b) is result

    # -- test methods ----------------

    def test_copy(self):
        """Test `Channel.copy()`."""
        new = self.TEST_CLASS(
            "X1:TEST",
            sample_rate=128,
            unit="m",
            frequency_range=(1, 40),
            safe=False,
            dtype="float64",
        )
        copy = new.copy()
        for attr in (
            "name",
            "ifo",
            "system",
            "subsystem",
            "signal",
            "trend",
            "type",
            "sample_rate",
            "unit",
            "dtype",
            "frametype",
            "model",
            "url",
            "frequency_range",
            "safe",
        ):
            a = getattr(new, attr)
            b = getattr(copy, attr)
            if isinstance(a, units.Quantity):
                utils.assert_quantity_equal(a, b)
            else:
                assert a == b

    @pytest.mark.parametrize(("name", "pdict"), [
        pytest.param(
            "X1:TEST-CHANNEL_NAME_PARSING.rms,m-trend",
            {
                "ifo": "X1",
                "system": "TEST",
                "subsystem": "CHANNEL",
                "signal": "NAME_PARSING",
                "trend": "rms",
                "type": "m-trend",
            },
            id="LIGO-m-trend",
        ),
        pytest.param(
            "G1:PSL_SL_PWR-AMPL-OUTLP-av",
            {
                "ifo": "G1",
                "system": "PSL",
                "subsystem": "SL",
                "signal": "PWR-AMPL-OUTLP",
                "trend": "av",
                "type": None,
            },
            id="GEO trend",
        ),
        pytest.param(
            "V1:h_16384Hz",
            {
                "ifo": "V1",
                "system": "h",
                "subsystem": "16384Hz",
                "signal": None,
                "trend": None,
                "type": None,
            },
            id="Virgo",
        ),
        pytest.param(
            "V1:Sa_PR_f0_zL_500Hz",
            {
                "ifo": "V1",
                "system": "Sa",
                "subsystem": "PR",
                "signal": "f0_zL_500Hz",
                "trend": None,
                "type": None,
            },
            id="Virgo2",
        ),
        pytest.param(
            "LVE-EX:X3_810BTORR.mean,m-trend",
            {
                "ifo": None,
                "system": "X3",
                "subsystem": "810BTORR",
                "signal": None,
                "trend": "mean",
                "type": "m-trend",
            },
            id="LIGO-LVE",
        ),
    ])
    def test_parse_channel_name(self, name, pdict):
        """Test `Channel.parse_channel_name()`."""
        # check empty parse via __init__
        c = self.TEST_CLASS("")
        for key in pdict:
            assert getattr(c, key) is None

        # check errors
        with pytest.raises(
            ValueError,
            match=r"^Cannot parse channel name according to LIGO-T990033$",
        ):
            self.TEST_CLASS.parse_channel_name("blah")

        # check parsing returns expected result
        assert self.TEST_CLASS.parse_channel_name(name) == pdict

        # check parsing translates to attributes
        c = self.TEST_CLASS(name)
        for key in pdict:
            assert getattr(c, key) == pdict[key]
        assert c.ndsname == name

    @pytest.mark.requires("ciecplib")
    def test_query(self, requests_mock):
        """Test `Channel.query`."""
        # build fake CIS response
        name = "X1:TEST-CHANNEL"
        results = [{
            "name": name,
            "units": "m",
            "datarate": 16384,
            "datatype": 4,
            "source": "X1MODEL",
            "displayurl": "https://cis.ligo.org/channel/123456",
        }]

        # apply the mock
        requests_mock.get(
            f"https://cis.ligo.org/api/channel/?q={name}",
            json=results,
        )

        # run the query
        chan = self.TEST_CLASS.query(
            name,
            idp="https://idp.example.com/profile/SAML2/SOAP/ECP",
            kerberos=False,
        )

        # check
        assert chan.name == name
        assert chan.unit == units.m
        assert chan.sample_rate == 16384 * units.Hz
        assert chan.dtype == numpy.dtype("float32")
        assert chan.model == "x1model"
        assert chan.url == "https://cis.ligo.org/channel/123456"

    @pytest.mark.requires("ciecplib")
    def test_query_notfound(self, requests_mock):
        """Test `Channel.query` handling of an empty response."""
        name = "X1:TEST-CHANNEL"
        requests_mock.get(
            f"https://cis.ligo.org/api/channel/?q={name}",
            json=[],
        )
        with pytest.raises(
            ValueError,
            match=fr"^No channels found matching '{name}'$",
        ):
            self.TEST_CLASS.query(
                name,
                idp="https://idp.example.com/profile/SAML2/SOAP/ECP",
                kerberos=False,
            )

    @pytest.mark.requires("nds2")
    @pytest.mark.usefixtures("nds2_connection")
    def test_query_nds2(self):
        """Test `Channel.query_nds2()`."""
        c = self.TEST_CLASS.query_nds2("X1:test", host="test")
        assert c.name == "X1:test"
        assert c.sample_rate == 16 * units.Hz
        assert c.unit == units.m
        assert c.dtype == numpy.dtype("float32")
        assert c.type == "raw"

    @pytest.mark.requires("nds2")
    @pytest.mark.usefixtures("nds2_connection")
    def test_query_nds2_error(self):
        """Test `Channel.query_nds2()` error handling."""
        with pytest.raises(
            ValueError,
            match="unique NDS2 channel match not found for 'Z1:test'",
        ):
            self.TEST_CLASS.query_nds2("Z1:test", host="test")

    @pytest.mark.requires("nds2")
    def test_from_nds2(self):
        """Test `Channel.from_nds2()`."""
        nds2c = mocks.nds2_channel("X1:TEST-CHANNEL", 64, "m")
        c = self.TEST_CLASS.from_nds2(nds2c)
        assert c.name == "X1:TEST-CHANNEL"
        assert c.sample_rate == 64 * units.Hz
        assert c.unit == units.m
        assert c.dtype == numpy.dtype("float32")
        assert c.type == "raw"

    @pytest.mark.requires("arrakis")
    def test_from_arrakis(self):
        """Test `Channel.from_arrakis()`."""
        import arrakis
        maker = arrakis.Channel(
            "X1:TEST-CHANNEL",
            numpy.dtype("int32"),
            128,
        )
        chan = self.TEST_CLASS.from_arrakis(maker)
        assert chan.name == maker.name
        assert chan.sample_rate == maker.sample_rate * units.Hz
        assert chan.dtype == maker.data_type


# -- ChannelList ---------------------

class TestChannelList:
    """Test `ChannelList`."""

    TEST_CLASS = ChannelList
    ENTRY_CLASS = Channel

    NAMES = (
        "X1:GWPY-CHANNEL_1",
        "X1:GWPY-CHANNEL_2",
        "X1:GWPY-CHANNEL_3",
    )
    SAMPLE_RATES = (1, 4, 8)

    @pytest.fixture
    @classmethod
    def instance(cls):
        """Create a new instance of the class under test."""
        return cls.TEST_CLASS([
            cls.ENTRY_CLASS(n, sample_rate=s) for
            n, s in zip(cls.NAMES, cls.SAMPLE_RATES, strict=True)
        ])

    def test_from_names(self):
        """Test `ChannelList.from_names()`."""
        cl = self.TEST_CLASS.from_names(*self.NAMES)
        assert cl == list(map(self.ENTRY_CLASS, self.NAMES))

        cl2 = self.TEST_CLASS.from_names(",".join(self.NAMES))
        assert cl == cl2

    def test_find(self, instance):
        """Test `ChannelList.find()`."""
        assert instance.find(self.NAMES[2]) == 2

    def test_find_error(self, instance):
        """Test `ChannelList.find()` error handling."""
        with pytest.raises(
            ValueError,
            match="blah",
        ):
            instance.find("blah")

    def test_sieve(self, instance):
        """Test `ChannelList.sieve()`."""
        cl = instance.sieve(name="GWPY-CHANNEL")
        assert cl == instance

        cl = instance.sieve(name="X1:GWPY-CHANNEL_2", exact_match=True)
        assert cl[0] is instance[1]

        cl = instance.sieve(name="GWPY-CHANNEL", sample_range=[2, 16])
        assert cl == instance[1:]

    @pytest.mark.requires("nds2")
    @pytest.mark.usefixtures("nds2_connection")
    def test_query_nds2(self):
        """Test `ChannelList.query_nds2()`."""
        # test query_nds2
        c = self.TEST_CLASS.query_nds2(
            ["X1:test", "Y1:test"],
            host="test",
        )
        assert len(c) == 2
        assert c[0].name == "X1:test"
        assert c[0].sample_rate == 16 * units.Hz

    @pytest.mark.requires("nds2")
    @pytest.mark.usefixtures("nds2_connection")
    def test_query_nds2_error(self):
        """Test `ChannelList.query_nds2` error handling."""
        # check errors
        assert len(
            self.TEST_CLASS.query_nds2(["Z1:test"], host="test"),
        ) == 0

    @pytest.mark.requires("nds2")
    @pytest.mark.usefixtures("nds2_connection")
    def test_query_nds2_availability(self):
        """Test `ChannelList.query_nds2_availability`."""
        avail = self.TEST_CLASS.query_nds2_availability(
            ["X1:test"],
            1000000000,
            1000000010,
            host="test",
        )
        utils.assert_dict_equal(
            avail,
            {"X1:test": SegmentList([Segment(1000000000, 1000000008)])},
            utils.assert_segmentlist_equal,
        )

    def test_read_write_omega_config(self, tmp_path):
        """Test `ChannelList.read(... format='omega-scan')`."""
        tmp = tmp_path / "config.ini"
        # write OMEGA_CONFIG to file and read it back
        with tmp.open("w") as f:
            f.write(OMEGA_CONFIG)
        cl = self.TEST_CLASS.read(tmp, format="omega-scan")
        assert len(cl) == 2
        assert cl[0].name == "L1:GDS-CALIB_STRAIN"
        assert cl[0].sample_rate == 4096 * units.Hertz
        assert cl[0].frametype == "L1_HOFT_C00"
        assert cl[0].params == {
            "channelName": "L1:GDS-CALIB_STRAIN",
            "frameType": "L1_HOFT_C00",
            "sampleFrequency": 4096,
            "searchTimeRange": 64,
            "searchFrequencyRange": (0, float("inf")),
            "searchQRange": (4, 96),
            "searchMaximumEnergyLoss": 0.2,
            "whiteNoiseFalseRate": 1e-3,
            "searchWindowDuration": 0.5,
            "plotTimeRanges": (1, 4, 16),
            "plotFrequencyRange": (),
            "plotNormalizedEnergyRange": (0, 25.5),
            "alwaysPlotFlag": 1,
        }
        assert cl[1].name == "L1:PEM-CS_SEIS_LVEA_VERTEX_Z_DQ"
        assert cl[1].frametype == "L1_R"

        # write omega config again using ChannelList.write and read it back
        # and check that the two lists match
        with tmp.open("w") as f:
            cl.write(f, format="omega-scan")
        cl2 = type(cl).read(tmp, format="omega-scan")
        assert cl == cl2

    def test_read_write_clf(self, tmp_path):
        """Test `ChannelList.read(... format='ini')`."""
        tmp = tmp_path / "config.clf"
        # write clf to file and read it back
        with tmp.open("w") as f:
            f.write(CLF)
        cl = ChannelList.read(tmp)
        assert len(cl) == 4
        a = cl[0]
        assert a.name == "H1:GDS-CALIB_STRAIN"
        assert a.sample_rate == 16384 * units.Hz
        assert a.frametype == "H1_HOFT_C00"
        assert a.frequency_range[0] == 4. * units.Hz
        assert a.frequency_range[1] == float("inf") * units.Hz
        assert a.safe is False
        assert a.params == {
            "qhigh": "150",
            "safe": "unsafe",
            "fidelity": "clean",
        }
        b = cl[1]
        assert b.name == "H1:ISI-GND_STS_HAM2_X_DQ"
        assert b.frequency_range[0] == .1 * units.Hz
        assert b.frequency_range[1] == 60. * units.Hz
        c = cl[2]
        assert c.name == "H1:ISI-GND_STS_HAM2_Y_DQ"
        assert c.sample_rate == 256 * units.Hz
        assert c.safe is False
        d = cl[3]
        assert d.name == "H1:ISI-GND_STS_HAM2_Z_DQ"
        assert d.safe is True
        assert d.params["fidelity"] == "glitchy"

        # write clf again using ChannelList.write and read it back
        # and check that the two lists match
        with tmp.open("w") as f:
            cl.write(f)
        cl2 = type(cl).read(tmp)
        assert cl == cl2



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.detector.units`."""

import pytest
from astropy import units

from ..units import parse_unit

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.mark.parametrize(("arg", "unit"), [
    pytest.param(None, None),
    pytest.param(units.m, units.m),
    pytest.param("meter", units.m),
    pytest.param("Volts", units.V),
    pytest.param("Meters/Second", units.m / units.s),
    pytest.param("Amp", units.ampere),
    pytest.param("MPC", units.megaparsec),
    pytest.param("degrees_C", units.Unit("Celsius")),
    pytest.param("DegC", units.Unit("Celsius")),
    pytest.param("degrees_F", units.Unit("Fahrenheit")),
    pytest.param("time", units.second),  # LIGO default time 'unit'
    pytest.param("Time (sec)", units.second),  # Virgo default time 'unit'
    pytest.param("Seconds", units.second),  # GWOSC default time 'unit'
])
def test_parse_unit(arg, unit):
    """Test `parse_unit()`."""
    assert parse_unit(arg, parse_strict="silent") == unit


def test_parse_unit_parse_strict_raise():
    """Test that `parse_unit` raises exceptions appropriately."""
    # check that errors get raise appropriately
    with pytest.raises(
        ValueError,
        match=r"^'metre' did not parse as unit.* Did you mean meter?",
    ):
        parse_unit("metre", parse_strict="raise")


def test_parse_unit_parse_strict_warn():
    """Test that `parse_unit` emits warnings appropriately."""
    # check that warnings get posted, and a custom NamedUnit gets returned
    with pytest.warns(
        units.UnitsWarning,
        match=r"^'metre' did not parse as gwpy unit.* Did you mean meter?",
    ):
        u = parse_unit("metre", parse_strict="warn")
    assert isinstance(u, units.UnrecognizedUnit)
    assert str(u) == "metre"

    # assert that a newly-created unit only gets created once
    u2 = parse_unit("metre", parse_strict="warn")
    assert u2 is u  # same object


@pytest.mark.parametrize("name", [
    "NONE",
    "undef",
    "strain",
    "coherence",
    "sec",
    "torr",
    "cf",
    "cfm",
    "ptcls",
])
def test_detector_units(name):
    """Test that a bunch of custom units are registered."""
    assert units.Unit(name)



# Copyright (c) 2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unified I/O read/write for `FrequencySeries` objects."""

from __future__ import annotations

from ..types.connect import (
    Array2DRead,
    Array2DWrite,
    SeriesRead,
    SeriesWrite,
)

# -- FrequencySeries -----------------

class FrequencySeriesRead(SeriesRead):
    """Read data into a `FrequencySeries`.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    source : `str`, `os.PathLike`, `file`, or `list` of these
        Source of data, any of the following:

        - Path of a single data file
        - List of data file paths
        - Path of LAL-format cache file

    args
        Other arguments are (in general) specific to the given ``format``.

    format : `str`, optional
        Source format identifier.
        If not given, the format will be detected if possible.
        See below for list of acceptable formats.

    kwargs
        Other keywords are (in general) specific to the given ``format``.

    Raises
    ------
    IndexError
        If ``source`` is an empty list.

    Notes
    -----"""


class FrequencySeriesWrite(SeriesWrite):
    """Write this `FrequencySeries` to a file.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    target : `str`
        Output filename.

    format : `str`, optional
        Output format identifier.
        If not given, the format will be detected if possible.
        See below for list of acceptable formats.

    Notes
    -----"""


# -- SpectralVariance ----------------

class SpectralVarianceRead(Array2DRead):
    """Read data into a `SpectralVariance`.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    source : `str`, `os.PathLike`, `file`, or `list` of these
        Source of data, any of the following:

        - Path of a single data file
        - List of data file paths
        - Path of LAL-format cache file

    args
        Other positional arguments are (in general) specific to the given ``format``.
        See below for list of acceptable formats.

    format : `str`, optional
        Source format identifier.
        If not given, the format will be detected if possible.
        See below for list of acceptable formats.

    kwargs
        Other keywords are (in general) specific to the given ``format``.
        See below for list of acceptable formats.

    Raises
    ------
    IndexError
        If ``source`` is an empty list.

    Notes
    -----"""


class SpectralVarianceWrite(Array2DWrite):
    """Write this `SpectralVariance` to a file.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format, the
    parameters below are common to most formats.

    Parameters
    ----------
    target : `str`
        Output filename.

    format : `str`, optional
        Output format identifier.
        If not given, the format will be detected if possible.
        See below for list of acceptable formats.

    Notes
    -----"""



# Copyright (c) 2018-2020 Louisiana State University
#               2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.signal.qtransform`."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy
import pytest
from numpy.random import default_rng
from numpy.testing import (
    assert_allclose,
)
from scipy.signal import gausspulse

from ...table import EventTable
from ...timeseries import TimeSeries
from .. import qtransform

if TYPE_CHECKING:
    from ...spectrogram import Spectrogram
    from ..qtransform import QGram

__author__ = "Alex Urban <alexander.urban@ligo.org>"

RNG = default_rng(seed=0)


# -- Fixtures -----------------------

SEARCH = (-0.25, 0.25)

@pytest.fixture(scope="module")
def noise() -> TimeSeries:
    """Generate a noise TimeSeries for testing."""
    return TimeSeries(
        RNG.normal(size=4096 * 10),
        sample_rate=4096,
        epoch=-5,
    )


@pytest.fixture(scope="module")
def glitch(noise) -> TimeSeries:
    """Generate a glitch TimeSeries for testing."""
    return TimeSeries(
        gausspulse(noise.times.value, fc=500) * 10,
        sample_rate=4096,
        epoch=-0.1,
    )


@pytest.fixture(scope="module")
def data(noise, glitch) -> TimeSeries:
    """Combine noise and glitch TimeSeries for testing."""
    return noise + glitch


@pytest.fixture(scope="module")
def qgram_far(data) -> tuple[QGram, float]:
    """Generate a Q-transform QGram and FAR for testing."""
    return qtransform.q_scan(data, search=SEARCH)


@pytest.fixture(scope="module")
def qgram(qgram_far) -> QGram:
    """Extract QGram from QGram and FAR tuple."""
    return qgram_far[0]


@pytest.fixture(scope="module")
def qfar(qgram_far) -> float:
    """Extract FAR from QGram and FAR tuple."""
    return qgram_far[1]


@pytest.fixture(scope="module")
def qspecgram(qgram) -> Spectrogram:
    """Interpolate a Q-transform QGram to a QSpecGram for testing."""
    return qgram.interpolate()


# -- Tests ---------------------------

def test_q_scan_gaussian(noise):
    """Test `qtransform.q_scan()` with a Gaussian signal."""
    qgram, far = qtransform.q_scan(noise, search=SEARCH)

    # Check FAR
    assert far == pytest.approx(0.029618458114325303)

    # Check that Q-plane frequencies are strictly increasing
    freq = qgram.plane.frequencies
    assert (freq[1:] > freq[:-1]).all()

    # Check peak
    peak = qgram.peak
    assert peak["energy"] == pytest.approx(13.86928)
    assert peak["frequency"] == pytest.approx(710.9)
    assert peak["snr"] == pytest.approx(5.266740798950195)


def test_q_scan_glitch(data):
    """Test `qtransform.q_scan()`."""
    qgram, far = qtransform.q_scan(data, search=SEARCH)

    # test that FAR is better than 1 / Hubble time
    assert far < 1 / (1.37e10 * 365 * 86400)

    # Check peak
    peak = qgram.peak
    assert peak["energy"] == pytest.approx(405.44434)
    assert peak["frequency"] == pytest.approx(505.6)
    assert peak["snr"] == pytest.approx(28.476107)


def test_timeseries_q_transform(data, qspecgram):
    """Test `TimeSeries.q_transform()` against `qtransform.q_scan()`."""
    # scan with the TimeSeries method
    ts_qspecgram = data.q_transform(whiten=False)

    # test spectrogram output
    assert ts_qspecgram.q == qspecgram.q
    assert ts_qspecgram.shape == qspecgram.shape
    assert ts_qspecgram.dtype == numpy.dtype("float32")
    assert_allclose(ts_qspecgram.value, qspecgram.value)


def test_timeseries_q_transform_unnormalised(data, qspecgram):
    """Test `TimeSeries.q_transform()` with norm=False."""
    # scan with norm=False
    ts_qspecgram = data.q_transform(whiten=False, norm=False)

    # test spectrogram output
    assert ts_qspecgram.q == qspecgram.q
    assert ts_qspecgram.dtype == numpy.dtype("float64")


def test_q_scan_fd(data, qfar, qspecgram):
    """Test frequency-domain `qtransform.q_scan()` against time-domain."""
    # create test object from frequency-domain input
    fs = data.fft()
    fs_qgram, far = qtransform.q_scan(
        fs,
        duration=abs(data.span),
        sampling=data.sample_rate.value,
        search=SEARCH,
        epoch=fs.epoch.value,
    )
    fs_qspecgram = fs_qgram.interpolate()

    # test that the output is the same
    assert far == qfar
    assert fs_qspecgram.q == qspecgram.q
    assert fs_qspecgram.dtype == numpy.dtype("float32")
    assert fs_qspecgram.shape == qspecgram.shape
    assert_allclose(fs_qspecgram.value, qspecgram.value, rtol=3e-2)


def test_qtable(qgram):
    """Test Q-transform QGram table output."""
    qtable = qgram.table()

    # Check basic properties
    assert isinstance(qtable, EventTable)
    assert qtable.meta["q"] == qgram.plane.q
    assert qtable["time"].shape == qtable["frequency"].shape

    # Check that peak properties are correctly reported
    imax = qtable["energy"].argmax()
    assert qtable["time"][imax] == qgram.peak["time"]
    assert qtable["duration"][imax] == 1 / 1638.4
    assert qtable["frequency"][imax] == qgram.peak["frequency"]
    assert qtable["bandwidth"][imax] == pytest.approx(
        2
        * numpy.pi ** (1 / 2.)
        * qtable["frequency"][imax]
        / qgram.plane.q,
    )
    assert qtable["energy"][imax] == qgram.peak["energy"]



@pytest.mark.parametrize(("snr", "size"), [
    pytest.param(0, 223232, id="0"),
    pytest.param(10, 37, id="10"),
    pytest.param(1e9, 0, id="high"),
])
def test_qtable_snrthresh(qgram, snr, size):
    """Test Q-transform QGram table output with SNR threshold."""
    # Test that too high an SNR threshold returns an empty table
    assert len(qgram.table(snrthresh=snr)) == size



# Copyright (c) 2017-2020 Joseph Areeda
#               2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Q-transform plots."""

from __future__ import annotations

import logging
import os.path
import re
from collections.abc import Sequence
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    cast,
)

from astropy.units import Quantity

from ..segments import Segment
from ..time import to_gps
from .spectrogram import SpectrogramProduct

if TYPE_CHECKING:
    from argparse import (
        ArgumentParser,
        Namespace,
        _ArgumentGroup,
    )
    from logging import Logger
    from os import PathLike
    from typing import (
        ClassVar,
        SupportsFloat,
    )

    from ..plot import Axes
    from ..spectrogram import Spectrogram

logger = logging.getLogger(__name__)


class QtransformProduct(SpectrogramProduct):
    """Plot the Q-transform (Omega)."""

    DEFAULT_FFTLENGTH = None  # type: ignore[assignment]
    MAX_DATASETS = 1
    action: ClassVar[str] = "qtransform"

    def __init__(
        self,
        args: Namespace,
        logger: Logger = logger,
    ) -> None:
        """Create a new `QtransformProduct`."""
        super().__init__(args, logger=logger)

        args = self.args
        self.qxfrm_args = {
            "gps": float(args.gps),
            "search": args.search / 2.,
            "fres": 0.5,
            "tres": args.tres,
            "whiten": not args.nowhiten,
        }
        if args.qrange is not None:
            self.qxfrm_args["qrange"] = args.qrange
        if args.frange is not None:
            self.qxfrm_args["frange"] = args.frange

    @classmethod
    def init_data_options(cls, parser: ArgumentParser) -> None:
        """Set up data input and signal processing options."""
        super().init_data_options(parser)
        cls.arg_qxform(parser)

    @classmethod
    def arg_channels(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for channel options."""
        group = parser.add_argument_group("Data options", "What data to load")
        group.add_argument(
            "--chan",
            required=True,
            help="Channel name.",
        )
        group.add_argument(
            "--gps",
            type=to_gps,
            required=True,
            help="Central time of transform",
        )
        group.add_argument(
            "--search",
            type=float,
            default=64,
            help="Time window around GPS to search",
        )
        return group

    @classmethod
    def arg_signal(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for signal options."""
        group = super().arg_signal(parser)
        group.add_argument(
            "--sample-freq",
            type=float,
            default=2048,
            help="Downsample freq",
        )
        return group

    @classmethod
    def arg_plot(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for plot options."""
        group = super().arg_plot(parser)

        # remove --out option
        outopt = next(act for act in group._actions if act.dest == "out")
        group._remove_action(outopt)

        # and replace with --outdir
        group.add_argument(
            "--outdir",
            default=os.path.curdir,
            dest="out",
            type=os.path.abspath,
            help="Directory for output images",
        )

        return group

    @classmethod
    def arg_qxform(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse.ArgumentGroup` for Q-transform options."""
        group = parser.add_argument_group("Q-transform options")
        group.add_argument(
            "--plot",
            nargs="+",
            type=float,
            default=[.5],
            help="One or more times to plot",
        )
        group.add_argument(
            "--frange",
            nargs=2,
            type=float,
            help="Frequency range to plot",
        )
        group.add_argument(
            "--qrange",
            nargs=2,
            type=float,
            help="Search Q range",
        )

        group.add_argument(
            "--nowhiten",
            action="store_true",
            help="do not whiten input before transform",
        )
        return group

    def _finalize_arguments(self, args: Namespace) -> None:
        """Derive standard args from our weird ones."""
        gps = args.gps
        search = args.search
        # ensure we have enough data for filter settling
        max_plot = max(args.plot)
        search = max(search, max_plot * 2 + 8)
        args.search = search
        self.logger.debug(
            "Search window: %s sec, max plot window %s",
            search,
            max_plot,
        )

        # make sure we don't create too big interpolations

        xpix = 1200.
        if args.geometry:
            m = re.match("(\\d+)x(\\d+)", args.geometry)
            if m:
                xpix = float(m.group(1))
        # save output x for individual tres calulation
        args.nx = xpix
        self.args.tres = search / xpix / 2
        self.logger.debug("Max time resolution (tres) set to %s", self.args.tres)

        args.start = [[int(gps - search/2)]]
        if args.epoch is None:
            args.epoch = args.gps
        args.duration = search

        args.chan = [[args.chan]]

        if args.color_scale is None:
            args.color_scale = "linear"

        args.overlap = 0  # so that FFTMixin._finalize_arguments doesn't fail

        xmin = args.xmin
        xmax = args.xmax

        super()._finalize_arguments(args)

        # unset defaults from `TimeDomainProduct`
        args.xmin = xmin
        args.xmax = xmax

    def get_ylabel(self) -> str:
        """Return the default y-axis label."""
        return "Frequency (Hz)"

    def get_color_label(self) -> str:
        """Return the default colorbar label."""
        return "Normalized energy"

    def get_suptitle(self) -> str:
        """Return the default super-title for plot."""
        return f"Q-transform: {self.chan_list[0]}"

    def get_title(self) -> str:
        """Return the default title for plot."""
        result = cast("Spectrogram", self.result)

        def fformat(x: SupportsFloat | Sequence[SupportsFloat]) -> str:
            """Format a float for display."""
            if isinstance(x, Sequence):
                return f"[{', '.join(map(fformat, x))}]"
            if isinstance(x, Quantity):
                x = x.value
            return f"{x:.2f}"

        bits: list[tuple[str, ...]] = [("Q", fformat(result.q))]
        bits.append(("tres", f"{self.qxfrm_args['tres']:.3g}"))
        if self.qxfrm_args.get("qrange"):
            bits.append(("q-range", fformat(self.qxfrm_args["qrange"])))
        if self.qxfrm_args["whiten"] is not None:
            bits.append(("whitened",))
        bits.extend([
            ("f-range", fformat(result.yspan)),
            ("e-range", f"[{result.min():.3g}, {result.max():.3g}]"),
        ])
        return ", ".join([": ".join(bit) for bit in bits])

    def get_spectrogram(self) -> Spectrogram | None:
        """Generate a Q-transform `Spectrogram`."""
        args = self.args

        fftlength = args.secpfft
        overlap = args.overlap * fftlength if fftlength else None
        method = args.method
        asd = self.timeseries[0].asd(
            fftlength=fftlength,
            overlap=overlap,
            method=method,
        )

        if (asd.value.min() == 0):
            self.logger.error(
                "Input data has a zero in ASD. Q-transform not possible.",
            )
            self.got_error = True
            return None

        gps = self.qxfrm_args["gps"]
        outseg = Segment(gps, gps).protract(args.plot[self.plot_num])

        # use the precomputed ASD as the whitener if needed
        if self.qxfrm_args.get("whiten") is True:
            self.qxfrm_args["whiten"] = asd

        # This section tries to optimize the amount of data that is
        # processed and the time resolution needed to create a good
        # image. NB:For each time span specified
        # NB: the timeseries h enough data for the longest plot
        inseg = outseg.protract(4) & self.timeseries[0].span
        proc_ts = self.timeseries[0].crop(*inseg)

        #  time resolution is calculated to provide about 4 times
        # the number of output pixels for interpolation
        tres = float(outseg.end - outseg.start) / 4 / self.args.nx
        self.qxfrm_args["tres"] = tres
        self.qxfrm_args["search"] = int(len(proc_ts) * proc_ts.dt.value)

        self.logger.debug("Q-transform arguments:")
        self.logger.debug("         outseg = %s", outseg)
        for key, val in sorted(self.qxfrm_args.items()):
            self.logger.debug("%15s = %s", key, val)

        qtrans = proc_ts.q_transform(outseg=outseg, **self.qxfrm_args)

        if args.ymin is None:  # set before Spectrogram.make_plot
            args.ymin = qtrans.yspan[0]

        return qtrans

    def _scale_axes_from_data(self, ax: Axes) -> None:
        """Scale axes based on data."""
        return super()._scale_axes_from_data(ax)

    def has_more_plots(self) -> bool:
        """Return `True` if there are more plots to make."""
        return self.plot_num < len(self.args.plot)

    def save(self, outdir: str | PathLike) -> None:  # type: ignore[override]
        """Save the plot to the specified output directory."""
        ts = self.timeseries[0]
        cname = re.sub(
            r"[-_:]", "_",
            str(ts.name or ts.channel or "DATA"),
        ).replace("_", "-", 1)
        png = f"{cname}-{float(self.args.gps)}-{self.args.plot[self.plot_num]}.png"
        outfile = Path(outdir) / png
        return super().save(outfile)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.utils`."""

import gzip
import tempfile
from pathlib import Path

import pytest

from .. import (
    cache as io_cache,
    utils as io_utils,
)
from .test_cache import cache  # noqa: F401

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def test_gopen(tmp_path):
    """Test `gopen` with a normal file."""
    tmp = tmp_path / "test.tmp"
    # test simple use
    with tmp.open("w") as f:
        f.write("blah blah blah")
    with pytest.deprecated_call():  # noqa: SIM117
        with io_utils.gopen(tmp) as f2:
            assert f2.read() == "blah blah blah"


@pytest.mark.parametrize("suffix", [".txt.gz", ""])
def test_gopen_gzip(tmp_path, suffix):
    """Test `gopen` with a gzip file."""
    tmp = tmp_path / f"test{suffix}"
    text = b"blah blah blah"
    with gzip.open(tmp, "wb") as fobj:
        fobj.write(text)
    with (
        pytest.deprecated_call(),
        io_utils.gopen(tmp, mode="rb") as fobj2,
    ):
        assert isinstance(fobj2, gzip.GzipFile)
        assert fobj2.read() == text


def test_file_list_file():
    """Test `file_list` with different file-like objects."""
    # test file -> [file.name]
    with tempfile.NamedTemporaryFile() as f:
        assert io_utils.file_list(f) == [f.name]


def test_file_list_cache(cache):  # noqa: F811
    """Test `file_list` with `lal.CacheEntry` objects."""
    lal_utils = pytest.importorskip("lal.utils")
    # test CacheEntry -> [CacheEntry.path]
    lcache = list(map(lal_utils.CacheEntry.from_T050017, cache))
    assert io_utils.file_list(lcache[0]) == [cache[0]]

    # test cache object -> pfnlist
    assert io_utils.file_list(lcache) == cache

    # test cache file -> pfnlist()
    with tempfile.NamedTemporaryFile(suffix=".lcf", mode="w") as f:
        io_cache.write_cache(lcache, f)
        f.seek(0)
        assert io_utils.file_list(f.name) == cache


@pytest.mark.parametrize("arg", [
    pytest.param("A,B,C,D", id="comma-separated"),
    pytest.param(["A", "B", "C", "D"], id="list"),
])
def test_file_list_str(arg):
    """Test `file_list` with different string inputs."""
    assert io_utils.file_list(arg) == ["A", "B", "C", "D"]


def test_file_list_error():
    """Test `file_list` with a bad input."""
    with pytest.raises(
        ValueError,
        match=r"^Could not parse input 1 as one or more file-like objects$",
    ):
        io_utils.file_list(1)  # type: ignore[arg-type]


@pytest.mark.parametrize(("input_", "expected"), [
    ("test.txt", "test.txt"),
    ("file:///test/path.txt", "/test/path.txt"),
    (Path("test.txt"), "test.txt"),
])
def test_file_path(input_, expected):
    """Test `file_path()`."""
    assert io_utils.file_path(input_) == expected


def test_file_path_file():
    """Check that :func:`gwpy.io.utils.file_path` can handle open files."""
    with tempfile.NamedTemporaryFile() as f:
        assert io_utils.file_path(f) == f.name


@pytest.mark.parametrize("badthing", [
    1,
    ["test.txt"],
])
def test_file_path_errors(badthing):
    """Check that :func:`gwpy.io.utils.file_path` fails when expected."""
    with pytest.raises(
        ValueError,
        match=r"^cannot parse file name for ",
    ):
        io_utils.file_path(badthing)


def test_file_path_cacheentry():
    """Check that :func:`gwpy.io.utils.file_path` can handle `CacheEntry`."""
    lal_utils = pytest.importorskip("lal.utils")
    path = "/path/to/A-B-0-1.txt"
    assert io_utils.file_path(lal_utils.CacheEntry.from_T050017(path)) == path



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for plotting on logarithmically-scaled axes."""

from __future__ import annotations

from functools import wraps
from math import (
    ceil,
    floor,
    isclose,
    log,
)
from typing import TYPE_CHECKING

import numpy
from matplotlib import (
    rcParams,
    ticker as mticker,
)
from matplotlib.scale import (
    LogScale as _LogScale,
    register_scale,
)

if TYPE_CHECKING:
    from matplotlib.axis import Axis
    from numpy.typing import NDArray

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def _math(s: str | float) -> str:
    """Format a number using mathtext or TeX."""
    if rcParams["text.usetex"]:
        return f"${s}$"
    return fr"$\mathdefault{{{s}}}$"


def _render_simple(
    values: NDArray,
    ndec: int = 2,
    thresh: float = 100001,
) -> bool:
    """Return `True` if all of ``values`` can be rendered as simple floats."""
    return (
        # is not empty
        bool(values.size)
        # and not all zeros
        and not all(values == 0)
        # and max below very large number
        and values.max() < thresh
        # and all <2 dec. places
        and numpy.array_equal(values, values.round(ndec))
    )


class LogFormatter(mticker.LogFormatterMathtext):
    """Format values for log axis.

    This `LogFormatter` extends the standard
    `~matplotlib.ticker.LogFormatterMathtext` to print numbers in the
    range [0.01, 1000) normally, and all others via the standard
    `~matplotlib.ticker.LogFormatterMathtext` output.
    """

    @wraps(mticker.LogFormatterMathtext.format_ticks)
    def format_ticks(self, values: list[float]) -> list[str]:
        """Format the set of ticks."""
        # this method overrides the default to enable formatting ticks
        # using simple float/integer representations (as opposed) to
        # scientific notation, if _all_ of the ticks have a value
        # small enough to render nicely (roughly <=1000 for visible ticks)
        # and no more than two decimal places
        self.set_locs(values)

        # remove floating-point precision errors
        values2 = numpy.asanyarray([float(f"{x:.12g}") for x in values])

        # if can render using just "%s" do it
        if _render_simple(values2):
            return [
                self(int(x) if x.is_integer() else x, pos=i, fmt="%s")
                for i, x in enumerate(values2)
            ]
        # otherwise use the matplotlib default
        return super().format_ticks(values)

    def _num_ticks(self) -> int:
        """Return the number of ticks that should be displayed."""
        if self.axis is None:
            return 0
        viewlim = self.axis.get_view_interval()
        loglim = numpy.log10(viewlim)
        return numpy.arange(
            ceil(loglim[0]),
            floor(ceil(loglim[1])),
            dtype=int,
        ).size

    @wraps(mticker.LogFormatterMathtext.set_locs)
    def set_locs(self, locs: list[float] | None = None) -> None:
        """Set the locations of the ticks."""
        ret = super().set_locs(locs=locs)

        # if a single major tick, but matplotlib decided not to include
        # sub-ticks, then there is more than a decade on the axis, so
        # (for base 10) include half-decade ticks
        if (
                self._num_ticks() == 1
                and self._sublabels == {1}
                and self._base == 10  # noqa: PLR2004
        ):
            self._sublabels = {1., 5., 10.}

        return ret

    @wraps(mticker.LogFormatterMathtext.__call__)
    def __call__(
        self,
        x: float,
        pos: int | None = None,
        fmt: str | None = None,
    ) -> str:
        """Format a tick label."""
        if x == 0:
            return _math((fmt or "%s") % 0)

        # work out whether to show this label
        # if there are enough major ticks or this formatter doesn't support
        # minor ticks, return a blank string
        x = abs(x)
        b = self._base
        fx = log(x) / log(b)
        is_x_decade = isclose(fx, round(fx))
        exponent = numpy.round(fx) if is_x_decade else numpy.floor(fx)
        coeff = numpy.round(x / b ** exponent)
        nticks = self._num_ticks()
        if (
                nticks >= 1
                and self._sublabels is not None
                and coeff not in self._sublabels
        ):
            return ""

        # enable custom format
        if fmt:
            sign = "-" if x < 0 else ""
            return _math(sign + fmt % x)

        return super().__call__(x, pos=pos)


class LogScale(_LogScale):
    """GWpy version of the matplotlib `LogScale`.

    This scale overrides the default to use the new GWpy formatters
    for major and minor ticks.
    """

    @wraps(_LogScale.set_default_locators_and_formatters)
    def set_default_locators_and_formatters(self, axis: Axis) -> None:
        """Set the default locators and formatters."""
        axis.set_major_locator(
            mticker.LogLocator(self.base),
        )
        axis.set_major_formatter(
            LogFormatter(self.base),
        )
        axis.set_minor_locator(
            mticker.LogLocator(self.base, self.subs),
        )
        axis.set_minor_formatter(
            LogFormatter(self.base, labelOnlyBase=bool(self.subs is not None)),
        )


register_scale(LogScale)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read LIGO_LW documents into :class:`~igwn_ligolw.ligolw.Table` objects."""

from __future__ import annotations

import inspect
from functools import cache
from typing import (
    TYPE_CHECKING,
    cast,
)

import numpy
from astropy.utils.compat.numpycompat import COPY_IF_NEEDED

from ...io.ligolw import (
    is_ligolw,
    read_table as read_ligolw_table,
    to_table_type as to_ligolw_table_type,
    write_tables as write_ligolw_tables,
)
from ...segments import Segment
from .. import (
    EventTable,
    Table,
)
from .utils import (
    dynamic_columns,
    read_with_where,
)

if TYPE_CHECKING:
    from collections.abc import (
        Collection,
        Iterable,
    )
    from typing import Any

    from astropy.table import Column
    from igwn_ligolw import ligolw
    from numpy.typing import (
        ArrayLike,
        DTypeLike,
    )

    from ...io.utils import (
        NamedReadable,
        Writable,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# methods to exclude from get_as_columns conversions
GET_AS_EXCLUDE = [
    "column",
    "username",
    "table",
    "time_slide_id",
]

# mappings
NUMPY_TYPE_MAP: dict[type, type] = {}
LIGOLW_TABLES: set[type] = set()
LIGOLW_PROPERTY_TYPES: tuple[type, ...] = (property,)

try:
    from igwn_ligolw import lsctables
except ImportError:
    HAVE_LSCTABLES = False
else:
    HAVE_LSCTABLES = True
    LIGOLW_TABLES = set(lsctables.TableByName.values())
    LIGOLW_PROPERTY_TYPES = (
        *LIGOLW_PROPERTY_TYPES,
        lsctables.gpsproperty,
        lsctables.gpsproperty_with_gmst,
        lsctables.instrumentsproperty,
        lsctables.segmentproperty,
    )
    NUMPY_TYPE_MAP[lsctables.LIGOTimeGPS] = numpy.float64


# -- utilities -----------------------

@cache
def _property_columns(prop: property) -> tuple[str, ...]:
    """Return the columns associated with a property descriptor."""
    from igwn_ligolw import lsctables
    if isinstance(prop, lsctables.gpsproperty_with_gmst):
        return (prop.s_name, prop.ns_name, prop.gmst_name)
    if isinstance(prop, lsctables.gpsproperty):
        return (prop.s_name, prop.ns_name)
    if isinstance(prop, lsctables.segmentproperty):
        return (prop.start, prop.stop)
    if isinstance(prop, lsctables.instrumentsproperty):
        return (prop.name,)
    return ()


def _get_property_columns(
    tabletype: type[ligolw.Table],
    columns: Iterable[str] | None,
) -> dict[str, set[str]]:
    """Return the `set` of columns required to read properties for a table.

    Examples
    --------
    >>> _get_property_columns(lsctables.SnglBurstTable, ['peak'])
    {'peak': {'peak_time', 'peak_time_ns'}}
    """
    rowtype = tabletype.RowType

    if columns is None:  # get all property columns
        try:
            columns = next(zip(
                *inspect.getmembers(
                    rowtype,
                    predicate=lambda x: isinstance(x, LIGOLW_PROPERTY_TYPES),
                ),
                strict=True,
            ))
        except StopIteration:
            # this object doesn't have any property columns
            return {}

    # get properties for row object
    rowvars = vars(rowtype)

    # build list of real column names for fancy properties
    extracols: dict[str, set[str]] = {}
    for key in columns:
        try:
            prop = rowvars[key]
        except KeyError:
            continue
        if propcols := _property_columns(prop):
            extracols.setdefault(key, set()).update(propcols)
    return extracols


@cache
def _get_property_type(
    tabletype: type[ligolw.Table],
    column: str,
) -> type | None:
    """Return the type of values in the given property.

    Examples
    --------
    >>> _get_property_type(lsctables.SnglBurstTable, 'peak')
    lal.LIGOTimeGPS
    """
    from igwn_ligolw import lsctables
    prop = vars(tabletype.RowType)[column]
    if isinstance(prop, lsctables.gpsproperty):
        return lsctables.LIGOTimeGPS
    if isinstance(prop, lsctables.segmentproperty):
        return Segment
    if isinstance(prop, lsctables.instrumentsproperty):
        return str
    return None


@cache
def table_columns(ligolw_table_class: type[ligolw.Table]) -> list[str]:
    """Return the list of column names for this table."""
    from igwn_ligolw import ligolw
    return list(map(ligolw.Column.ColumnName, ligolw_table_class.validcolumns))


# -- conversions ---------------------

def to_astropy_table(
    llwtable: ligolw.Table,
    table_class: type[Table],
    copy: bool | None = COPY_IF_NEEDED,  # noqa: FBT001
    *,
    columns: Collection[str] | None =None,
    use_numpy_dtypes: bool = False,
    rename: dict[str, str] | None = None,
) -> Table:
    """Convert a `~igwn_ligolw.ligolw.Table` to an `~astropy.tableTable`.

    This method is designed as an internal method to be attached to
    :class:`~igwn_ligolw.ligolw.Table` objects as `__astropy_table__`.

    Parameters
    ----------
    llwtable : `~igwn_ligolw.ligolw.Table`
        The LIGO_LW table to convert from.

    table_class : `type`
        Table type to convert to.

    copy : `bool`, optional
        If `True` copy the input data, otherwise return a reference.
        Default is `False`

    columns : `list` of `str`, optional
        The columns to populate.
        Defaults to all columns.

    use_numpy_dtypes : `bool`, optional
        If `True` then force all columns in the output to have standard
        Numpy types.

    rename : `dict`, optional
        Mapping of ``('old name' -> 'new name')`` pairs to rename columns
        from the original ``LIGO_LW`` table.

    Returns
    -------
    table : `Table`
        The newly transformed table. An instance of ``table_class``.
    """
    # set default keywords
    if rename is None:
        rename = {}
    if columns is None:
        columns = llwtable.columnnames

    # extract columns from LIGO_LW table as astropy.table.Column
    data = []
    for colname in columns:
        arr = _get_column(llwtable, colname)

        # transform to astropy.table.Column
        copythis = copy or isinstance(arr, numpy.ndarray)
        data.append(to_astropy_column(
            arr,
            table_class.Column,
            copy=copythis,
            use_numpy_dtype=use_numpy_dtypes,
            name=rename.get(colname, colname),
        ))

    # build table and return
    return table_class(
        data,
        copy=copy,
        meta={"tablename": str(llwtable.Name)},
    )


def _get_column(
    llwtable: ligolw.Table,
    name: str,
) -> ligolw.Column:
    """Get a column from a ``LIGO_LW`` table."""
    # handle empty tables as special (since we can't introspect the types)
    if not len(llwtable):
        dtype: type | None
        if name in llwtable.validcolumns:
            dtype = _get_pytype(llwtable.validcolumns[name])
        else:
            dtype = _get_property_type(type(llwtable), name)  # type: ignore[arg-type]
        if dtype:
            return numpy.empty((0,), dtype=dtype)

    try:
        # try normal column access
        return llwtable.getColumnByName(name)
    except KeyError:  # bad name
        # try get_{}
        get_ = f"get_{name}"
        if (
            hasattr(llwtable, get_)
            and name not in GET_AS_EXCLUDE
        ):
            return getattr(llwtable, get_)()

        # try array of property values
        try:
            return numpy.asarray([getattr(row, name) for row in llwtable])
        except AttributeError:  # no property
            pass

        raise  # KeyError from getColumnByName


def to_astropy_column(
    llwcol: ligolw.Column,
    cls: type[Column],
    *,
    copy: bool | None = COPY_IF_NEEDED,
    dtype: DTypeLike | None = None,
    use_numpy_dtype: bool = False,
    **kwargs,
) -> Column:
    """Convert a :class:`~igwn_ligolw.ligolw.Column` to `astropy.table.Column`.

    Parameters
    ----------
    llwcol : `~igwn_ligolw.ligolw.Column`, `numpy.ndarray`, iterable
        The ``LIGO_LW`` column to convert, or an iterable.

    cls : `~astropy.table.Column`
        The Astropy `~astropy.table.Column` or subclass to convert to.

    copy : `bool`, optional
        If `True` copy the input data, otherwise return a reference.
        Default is `False`.

    dtype : `type`, optional
        The data type to convert to when creating the `~astropy.table.Column`.

    use_numpy_dtype : `bool`, optional
        If `True` then force the output column to only have standard
        Numpy types.

    kwargs
        Other keyword arguments are passed to the `~astropy.table.Column`
        creator.

    Returns
    -------
    column : `~astropy.table.Column`
        An Astropy version of the given ``LIGO_LW`` column.
        An instance of ``cls``.
    """
    if dtype is None:  # try and find dtype
        dtype = _get_column_dtype(llwcol)
        if (
            use_numpy_dtype
            and numpy.dtype(dtype).type is numpy.object_
        ):
            dtype = cast("type", dtype)
            # dtype maps to 'object' in numpy, try and resolve real numpy type
            try:
                dtype = NUMPY_TYPE_MAP[dtype]
            except KeyError as exc:
                msg = f"no mapping from object type '{dtype}' to numpy type"
                raise TypeError(msg) from exc
    return cls(
        data=llwcol,
        copy=copy,
        dtype=dtype,
        **kwargs,
    )


def _get_column_dtype(
    llwcol: ligolw.Column | ArrayLike,
) -> type | None:
    """Get the data type of a LIGO_LW `Column`.

    Parameters
    ----------
    llwcol : `~igwn_ligolw.ligolw.Column`, `numpy.ndarray`, iterable
        A ``LIGO_LW`` column, a numpy array, or an iterable.

    Returns
    -------
    dtype : `type`, `None`
        The object data type for values in the given column, `None` is
        returned if ``llwcol`` is a `numpy.ndarray` with `numpy.object_`
        dtype, or no data type can be parsed (e.g. empty list).

    Examples
    --------
    >>> _get_column_dtype(numpy.zeros(1))
    dtype('float64')
    >>> _get_column_dtype(llwtable.getColumnByName('peak_time')
    'int32'
    """
    try:
        dtype = llwcol.dtype  # type: ignore[union-attr]
        if dtype is numpy.dtype("O"):
            raise AttributeError  # goto below
    except AttributeError:
        try:  # igwn_ligolw.ligolw.Column
            name = str(llwcol.getAttribute("Name"))  # type: ignore[union-attr]
            if name.startswith(f"{llwcol.parentNode.Name}:"):  # type: ignore[union-attr]
                name = name.split(":", 1)[-1]
            llwtype = llwcol.parentNode.validcolumns[name]  # type: ignore[union-attr]
        except AttributeError:  # not a column
            try:
                return type(llwcol[0])  # type: ignore[index]
            except IndexError:
                return None
        # map column type str to python type
        return _get_pytype(llwtype)
    return dtype


def _get_pytype(llwtype: type) -> type:
    """Return a dtype-compatible type for the given LIGO_LW type string."""
    from igwn_ligolw.types import (
        ToNumPyType,
        ToPyType,
    )
    try:
        return numpy.dtype(ToNumPyType[llwtype]).type
    except KeyError:
        return ToPyType[llwtype]


def table_to_ligolw(
    table: Table,
    tablename: str,
) -> ligolw.Table:
    """Convert a `astropy.table.Table` to a `igwn_ligolw.ligolw.Table`."""
    from igwn_ligolw import lsctables

    # -- work out columns for LIGO_LW table
    # this is overly complicated because of the way that we magically
    # map properties of igwn_ligolw.ligolw.Table objects to real columns in
    # astropy.table.Table objects, but also because igwn_ligolw internally
    # combines table names and column names for some columns
    # (e.g. process:process_id)
    columns = table.columns.keys()
    cls = lsctables.TableByName[tablename]
    inst = cls.new()
    try:
        columnnamesreal = dict(zip(
            inst.columnnames,
            inst.columnnamesreal,
            strict=False,
        ))
    except AttributeError:  # glue doesn't have these attributes
        columnnamesreal = {}
    llwcolumns = [columnnamesreal.get(n, n) for n in columns]
    for col, llwcols in _get_property_columns(cls, columns).items():
        idx = llwcolumns.index(col)
        llwcolumns.pop(idx)
        for name in list(llwcols)[::-1]:
            llwcolumns.insert(idx, name)

    # create new LIGO_LW table
    llwtable = cls.new(columns=llwcolumns)

    # map rows across
    for row in table:
        llwrow = llwtable.RowType()
        for name in columns:
            setattr(llwrow, name,
                    to_ligolw_table_type(row[name], llwtable, name))
        llwtable.append(llwrow)

    return llwtable


# -- read ----------------------------

@read_with_where
def read_table(
    source: NamedReadable | ligolw.Document | list[NamedReadable],
    tablename: str | None = None,
    columns: Collection[str] | None = None,
    ligolw_columns: Collection[str] | None = None,
    **kwargs,
) -> Table:
    """Read a `Table` from one or more LIGO_LW XML documents.

    source : `file`, `str`, `~igwn_ligolw.ligolw.Document`, `list`
        One or more open files or file paths, or a single LIGO_LW ``Document``.

    tablename : `str`, optional
        The `Name` of the relevant ``LIGO_LW`` `Table` to read.
        Required if the document doesn't contain exactly one table.

    columns : `list` of `str`, optional
        The list of columns to include in the returned Table.
        Default is all columns.
        This list may include:

        - any column given in the ``validcolumns`` mapping
          for the table (in :mod:`igwn_ligolw.lsctables`)
        - any of the properties defined on the associated ``RowType`` object

    ligolw_columns : `list` of `str`, optional
        The list of columns to read from the ``LIGO_LW`` document.
        Default is all columns.
        This should be a super-set of ``columns``, including any
        ``validcolumns`` for the relevant table that are required to resolve
        dynamic columns from a row `property`.

    use_numpy_dtypes : `bool`, optional
        If `True` then force all columns in the output to have standard
        Numpy types.

    rename : `dict`, optional
        Mapping of ``('old name' -> 'new name')`` pairs to rename columns
        from the original ``LIGO_LW`` table.

    copy : `bool`, optional
        If `True` copy the input data, otherwise return a reference.
        Default is `False`

    See Also
    --------
    gwpy.io.ligolw.read_table
        For details of keyword arguments for the read operation.

    gwpy.table.io.ligolw.to_astropy_table
        For details of keyword arguments for the conversion operation.
    """
    from igwn_ligolw import ligolw
    from igwn_ligolw.lsctables import TableByName

    # -- keyword handling ------------

    # separate keywords for reading from LIGO_LW and converting to Astropy
    convert_kw: dict[str, Any] = {
        "rename": None,
        "use_numpy_dtypes": False,
    }
    for key in filter(kwargs.__contains__, convert_kw):
        convert_kw[key] = kwargs.pop(key)
    if convert_kw["rename"] is None:
        convert_kw["rename"] = {}
    read_kw = kwargs  # rename for readability

    # allow user to specify LIGO_LW columns to read to provide the
    # desired output columns
    read_kw["columns"] = ligolw_columns or columns
    convert_kw["columns"] = columns or read_kw["columns"]

    if tablename:
        tableclass = TableByName[ligolw.Table.TableName(tablename)]
        read_kw["columns"] = dynamic_columns(
            convert_kw["columns"],
            table_columns(tableclass),
            _get_property_columns(tableclass, None),
        )[0]

    # -- read ------------------------

    return Table(
        read_ligolw_table(
            source,
            tablename=tablename,
            **read_kw,
        ),
        **convert_kw,
    )


# -- write ---------------------------

def write_table(  # noqa: D417
    table: Table,
    target: Writable | ligolw.Document,
    tablename: str | None = None,
    **kwargs,
) -> None:
    """Write a `~astropy.table.Table` to file in LIGO_LW XML format.

    Parameters
    ----------
    target : `str`, `pathlib.Path`, `file`
        The file path or file object to write to.

    tablename : `str`, optional
        The Name of the ``LIGO_LW`` table to write to.
        Default is taken from the ``tabename`` metadata attribute of
        the table being written, otherwise this argument is required.

    kwargs
        All keyword arguments are passed to
        `gwpy.io.ligolw.write_tables`.

    See Also
    --------
    gwpy.io.ligolw.write_tables
        For details of the table writing implementation and any
        valid keyword arguments.
    """
    if tablename is None:  # try and get tablename from metadata
        tablename = table.meta.get("tablename", None)
    if tablename is None:  # panic
        msg = (
            "please pass ``tablename=`` to specify the target "
            "LIGO_LW Table Name"
        )
        raise ValueError(msg)
    llwtable = table_to_ligolw(
        table,
        tablename,
    )
    return write_ligolw_tables(target, [llwtable], **kwargs)


# -- register ------------------------

for table_ in LIGOLW_TABLES:
    # register conversion from LIGO_LW to astropy Table
    table_.__astropy_table__ = to_astropy_table  # type: ignore[attr-defined]

for klass in (Table, EventTable):
    registry = klass.read.registry
    registry.register_reader("ligolw", klass, read_table)
    registry.register_writer("ligolw", klass, write_table)
    registry.register_identifier("ligolw", klass, is_ligolw)



# Copyright (c) 2018-2020 Louisiana State University
#               2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. currentmodule:: gwpy.timeseries
.. sectionauthor:: Alex Urban <alexander.urban@ligo.org>

Inject a signal into a `FrequencySeries`
########################################

It can often be useful to add some known signal to inherently random
or noisy data. For example, one might want to investigate what
would happen if a binary black hole merger signal occured at or near
the time of a glitch. In LIGO data analysis, this procedure is referred
to as an *injection*.

In the example below we will create a stream of random, white Gaussian
noise, then inject a loud, steady sinuosoid. We will do this in the
frequency domain because it is much easier to model a sinusoid there.
"""

# %%
# Generate random data
# --------------------
# First, we prepare one second of Gaussian noise using
# `numpy.random.Generator.normal`:

from numpy.random import default_rng

from gwpy.timeseries import TimeSeries

rng = default_rng()
noise = TimeSeries(rng.normal(scale=.1, size=1024), sample_rate=1024)

# %%
# To inject a signal in the frequency domain, we need to take an FFT:

noisefd = noise.fft()

# %%
# Inject noise
# ------------
# We can now easily inject a loud sinusoid of unit amplitude at, say,
# 30 Hz. To do this, we use
# :meth:`~gwpy.frequencyseries.FrequencySeries.inject`:

import numpy

from gwpy.frequencyseries import FrequencySeries

signal = FrequencySeries(numpy.array([1.]), f0=30, df=noisefd.df)
injfd = noisefd.inject(signal)

# %%
# Visualisation
# -------------
# We can then visualize the data before and after injection in the frequency
# domain:

from gwpy.plot import Plot

plot = Plot(
    numpy.abs(noisefd),
    numpy.abs(injfd),
    xscale="log",
    yscale="log",
)
plot.show()

# %%
# Finally, for completeness we can visualize the effect before and after
# injection back in the time domain:

inj = injfd.ifft()
plot = Plot(
    noise,
    inj,
    separate=True,
    sharex=True,
    sharey=True,
    figsize=(12, 6),
)
plot.show()



# Copyright (c) 2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

r"""Create a reduced data set (RDS) using GWpy.

This tool allows you to create a reduced data set (RDS) by specifying
a start and end time, a list of channels, and an output file.
The tool will fetch the data for the specified channels and time range,
and write it to the output file in the specified format.
"""

from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING

from .. import __version__
from ..io.registry import default_registry
from ..time import to_gps
from ..timeseries import TimeSeriesDict
from . import _utils

if TYPE_CHECKING:
    from argparse import ArgumentParser
    from collections.abc import Sequence

    from ..time import SupportsToGps

logger = _utils.get_logger(__name__)

EXAMPLES = {
    "Get data for GW150914": " ".join((  # noqa: FLY002
        "gwpy-rds",
        "1126259462",
        "1126259522",
        "H1:GWOSC-4KHZ_R1_STRAIN",
        "L1:GWOSC-4KHZ_R1_STRAIN",
        "-o gw150914.gwf",
        "-O version=4",
    )),
}

DOC_VERSION = "latest" if "dev" in __version__ else __version__
DOC_EPILOG = rf"""
For more information, see the online documentation at:

https://gwpy.readthedocs.io/en/{DOC_VERSION}/tools/rds/
"""


def create_rds(
    channels: Sequence[str],
    start: SupportsToGps,
    end: SupportsToGps,
    outfile: Path,
    source: Sequence[str] | str | None = None,
    write_format: str | None = None,
    **kwargs,
) -> None:
    """Create a reduced data set by grabbing data and writing to a file.

    Parameters
    ----------
    channels : `list` of `str`
        The list of channel names to get.

    start : `LIGOTimeGPS`
        The GPS start time for the RDS.

    end : `LIGOTimeGPS`
        The GPS end time for the RDS.

    outfile : `str`, `pathlib.Path`
        The path of the output file to create.

    source : `str`, `list` of `str`, optional
        One or more supported data sources to use.

    write_format : `str`, optional
        The format option to use when writing data.
        Default is inferred from the output file path.
        See `TimeSeriesDict.write.help()` for details on supported formats.

    kwargs
        Other keyword arguments are passed to `TimeSeriesDict.get`.

    See Also
    --------
    TimeSeriesDict.get
        For details of how data are accessed.

    TimeSeriesDict.write
        For details of how data are written.
    """
    logger.info("Getting data")
    data = TimeSeriesDict.get(
        channels,
        start,
        end,
        source=source,
        **kwargs,
    )
    logger.info("Received data for %d channels", len(data))
    data.write(
        outfile,
        format=write_format,
        overwrite=True,
    )
    logger.info("Wrote %s", outfile)


# -- Docstring helpers ---------------

def _format_write_formats(epilog: str) -> tuple[str, list[dict[str, str]]]:
    """Document the list available formats for writing `TimeSeriesDict`."""
    # Get the list of available formats
    formats = [
        fmt["Format"]
        for fmt in default_registry.get_formats(TimeSeriesDict, readwrite="write")
    ]

    # Format the documentation for manual or help output
    formats_header = "Supported -f/--format values:"
    formats_lines = [formats_header, ""]
    man = _utils.is_manpage()
    for fmt in formats:
        if man:
            formats_lines.extend((
                r".IP \[bu]",
                fr"\fB{fmt}\fR",
            ))
        else:
           formats_lines.append(f"  - {fmt}")
    formats_doc = os.linesep.join(formats_lines)

    # Attach to epilog or manpage sections
    manpage_sections = []
    if man:
        manpage_sections.append(
            {
                "heading": "data formats",
                "content": formats_doc,
            },
        )
        return epilog, manpage_sections
    return formats_doc + os.linesep * 2 + epilog, []


# -- Command line interface ----------

def create_parser() -> ArgumentParser:
    """Create an `argparse.ArgumentParser` for this tool."""
    # Format the help message, including data sources
    epilog, manpage_sections = _format_write_formats(DOC_EPILOG)

    # Create parser
    parser = _utils.ArgumentParser(
        description=__doc__,
        epilog=epilog,
        examples=EXAMPLES,
        prog="gwpy-rds",
        manpage=manpage_sections,
    )
    parser.add_argument(
        "start",
        type=to_gps,
        help=(
            "Start time of data request. "
            "Can be specified as a GPS time, date/time string, or relative time; "
            "please ensure that date strings containing spaces are quoted."
        ),
    )
    parser.add_argument(
        "end",
        type=to_gps,
        help=(
            "End time of data request. "
            "Can be specified as a GPS time, date/time string, or relative time; "
            "please ensure that date strings containing spaces are quoted."
        ),
    )
    parser.add_argument(
        "channels",
        nargs="+",
        metavar="channel|ifo",
        help=(
            "Data channel or IFO to request; can be specified multiple times. "
            "An IFO prefix (e.g., 'H1') can be passed to request public strain "
            "data from GWOSC."
        ),
    )
    parser.add_argument(
        "-g",
        "--source",
        action="append",
        help=(
            "Source from which to get data. "
            "See `help(TimeSeries.get)` for documentation on the "
            "supported sources."
        ),
    )
    parser.add_argument(
        "-o",
        "--output-file",
        default="gwpy-rds.h5",
        type=Path,
        help="Output file in which to write data.",
    )
    parser.add_argument(
        "-f",
        "--format",
        help=(
            "Format in which to write output data. "
            "Default is inferred from -o/--output-file. "
            "See below or `help(TimeSeriesDict.write)` for supported formats."
        ),
    )
    parser.add_argument(
        "-O",
        "--option",
        action="append",
        default=None,
        metavar="key=value",
        help="Additional options to pass to the TimeSeriesDict.get.",
    )
    parser.add_argument(
        "-L",
        "--log-name",
        default="gwpy",
        help=(
            "Name of the logger to configure for verbose output; "
            "use 'root' to enable logging in all modules."
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity; pass once for INFO, twice for DEBUG.",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=__version__,
        help="Show the version number and exit.",
    )
    return parser


def main(args: list[str] | None = None) -> None:
    """Run the command line tool to create a reduced data set (RDS)."""
    # Parse the command line arguments
    parser = create_parser()
    opts = parser.parse_args(args=args)

    # Init verbose logging
    _utils.init_verbose_logging(opts.log_name, opts.verbose)

    # Parse additional options
    kwargs = _utils.parse_options_dict(opts.option or [])

    # Create the RDS
    create_rds(
        opts.channels,
        opts.start,
        opts.end,
        opts.output_file,
        opts.source,
        write_format=opts.format,
        **kwargs,
    )


if __name__ == "__main__":
    main()



# Copyright (c) 2019-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilties for enumerations."""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

import numpy

if TYPE_CHECKING:
    import builtins
    from typing import Self

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


class NumpyTypeEnum(Enum):
    """`~enum.Enum` of numpy types."""

    @property
    def dtype(self) -> numpy.dtype:
        """The numpy dtype corresponding to this enumerated type."""
        return numpy.dtype(self.name.lower())

    @property
    def type(self) -> builtins.type:
        """The python type corresponding to this enumerated type."""
        return self.dtype.type

    @classmethod
    def find(
        cls,
        type_: builtins.type | str | int,
    ) -> Self:
        """Return the enumerated type corresponding to the given python type."""
        try:
            return cls(type_)
        except ValueError as exc:
            if isinstance(type_, str):
                type_ = type_.lower()
            try:
                return cls[numpy.dtype(type_).name.upper()]  # type: ignore[arg-type]
            except (
                KeyError,  # numpy dtype isn't support by this enum
                TypeError,  # type isn't a valid numpy type
            ):
                raise exc from None



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilties for testing tables."""

from __future__ import annotations

from typing import TYPE_CHECKING

from astropy.table import Table
from numpy import dtype
from numpy.random import default_rng

from .. import (
    EventTable,
    GravitySpyTable,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from numpy.typing import (
        ArrayLike,
        DTypeLike,
    )

TABLE_CLASSES = (
    Table,
    EventTable,
    GravitySpyTable,
)


def random_table(
    names: Iterable[str],
    length: int = 100,
    table_class: type = EventTable,
    dtypes: list[DTypeLike] | None = None,
) -> Table:
    """Create a table full of random data.

    Parameters
    ----------
    names : `list` of `str`
        The names of the columns to create.

    length : `int`, optional
        The number of rows to create. Default is 100.

    table_class: `type`, optional
        The type of table to create, should be `~astropy.table.Table`
        or a sub-class.
        Default is `~gwpy.table.EventTable`.

    dtypes : `list` of `numpy.dtype`, `type`, `str, optional
        The list of types to associate with each column.
        If given the number of types must match the number of columns
        (``len(dtypes) == len(names)``).
        Defaults to `None` to return `numpy.float64` columns.

    Returns
    -------
    table : instance of ``table_class``
        A table filled with random data.
    """
    names = list(names)
    data: list[ArrayLike | Iterable] = []
    for i in range(len(names)):
        # generate data
        rng = default_rng(seed=i)
        col = rng.random(size=length) * 1000
        # cast to type
        if dtypes:
            dtp = dtypes[i]
            # use map() to support non-primitive types
            if dtype(dtp).name == "object" and callable(dtp):
                data.append(list(map(dtp, col)))
            else:
                data.append(col.astype(dtp))
        else:
            data.append(col)
    return table_class(
        data,
        names=names,
    )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.utils.misc`."""

import pytest

from .. import misc as utils_misc


def test_round_to_power():
    """Test for :func:`gwpy.utils.misc.round_to_power`."""
    # test basic features
    assert utils_misc.round_to_power(2) == 2
    assert utils_misc.round_to_power(9, base=10) == 10
    assert utils_misc.round_to_power(5, which="lower") == 4
    assert utils_misc.round_to_power(5, which="upper") == 8
    # test output
    base = 10.
    rounded = utils_misc.round_to_power(9, base=base)
    assert base == rounded
    assert type(base) == type(rounded)  # noqa: E721


def test_round_to_power_error():
    """Test for an errored use case of :func:`gwpy.utils.misc.round_to_power`."""
    with pytest.raises(
        ValueError,
        match=r"^'which' argument must be one of 'lower', 'upper', or None$",
    ):
        utils_misc.round_to_power(7, which="")


def test_unique():
    """Test for :func:`gwpy.utils.misc.unique`."""
    a = [1, 2, 4, 3, 5, 4, 5, 3]
    assert utils_misc.unique(a) == [1, 2, 4, 3, 5]


@pytest.mark.parametrize(("func", "value", "out"), [
    pytest.param(str, None, None, id="None"),
    pytest.param(str, 1, "1", id="str->int"),
])
def test_if_not_none(func, value, out):
    """Test for :func:`gwpy.utils.misc.if_not_none`."""
    assert utils_misc.if_not_none(func, value) == out



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Wrapper to the nds2-client package, providing network access to LIGO data."""

from __future__ import annotations

import enum
import logging
import operator
import os
import re
import warnings
from contextlib import contextmanager
from functools import (
    reduce,
    wraps,
)
from typing import (
    TYPE_CHECKING,
    cast,
)

from ..time import to_gps
from ..utils.enum import NumpyTypeEnum
from .kerberos import kinit

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
        Iterator,
        Sequence,
    )
    from typing import (
        ParamSpec,
        Self,
        TypeVar,
    )

    import nds2
    from numpy.typing import DTypeLike

    from ..detector import Channel
    from ..segments import SegmentListDict
    from ..time import SupportsToGps

    # Type variables for function decorators
    P = ParamSpec("P")
    T = TypeVar("T")

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = [
    "NDSWarning",
    "Nds2ChannelType",
    "Nds2DataType",
    "auth_connect",
    "connect",
    "find_channels",
    "get_availability",
    "host_resolution_order",
    "minute_trend_times",
    "parse_nds_env",
]

logger = logging.getLogger(__name__)

try:
    import nds2
except ModuleNotFoundError:
    MIN_SAMPLE_RATE = 0.
    MAX_SAMPLE_RATE = 1e12
else:
    MIN_SAMPLE_RATE = nds2.channel.MIN_SAMPLE_RATE
    MAX_SAMPLE_RATE = nds2.channel.MAX_SAMPLE_RATE

# regular expression to match LIGO-standard NDS1 hostnames (e.g x1nds0)
NDS1_HOSTNAME = re.compile(r"[a-z]1nds[0-9]\Z")

# map of default hosts for each interferometer prefix
DEFAULT_HOSTS = {
    None: ("nds.ligo.caltech.edu", None),
    "H1": ("nds.ligo-wa.caltech.edu", None),
    "H0": ("nds.ligo-wa.caltech.edu", None),
    "L1": ("nds.ligo-la.caltech.edu", None),
    "L0": ("nds.ligo-la.caltech.edu", None),
    "V1": ("nds.ligo.caltech.edu", None),
    "C1": ("nds40.ligo.caltech.edu", None),
    "C0": ("nds40.ligo.caltech.edu", None),
}
GWOSC_NDS2_HOSTS = [
    ("nds.gwosc.org", None),
]


# -- enums ---------------------------

class _Nds2Enum(enum.IntFlag):
    """Base class for NDS2 enums."""

    # NDS2 name for this type
    nds2name: str

    def __new__(
        cls,
        value: int,
        nds2name: str,
    ) -> Self:
        obj = int.__new__(cls, value)
        obj._value_ = value
        obj.nds2name = nds2name
        return obj

    @classmethod
    def any(cls) -> Self:
        """Return the logical OR of all members in this enum."""
        return reduce(operator.or_, cls)

    @classmethod
    def nds2names(cls) -> list[str]:
        """Return the list of all recognised NDS2 names for this type."""
        return [x.nds2name for x in cls]

    @classmethod
    def find(cls, name: str | int) -> Self:
        """Return the NDS2 type corresponding to the given name."""
        name = str(name)
        # if given a number, use it
        if name.isdigit():
            return cls(int(name))  # type: ignore[call-arg]
        # otherwise we might have been given a registered name
        try:
            return cls[name.upper()]
        except KeyError:
            # otherwise otherwise check the NDS2 names
            for item in cls:
                if name.lower() == item.nds2name.lower():
                    return item
        return cls["UNKNOWN"]


class Nds2ChannelType(_Nds2Enum):
    """NDS2 channel type with descriptive name."""

    UNKNOWN = (0, "UNKNOWN")
    ONLINE = (1, "online")
    RAW = (2, "raw")
    RDS = (4, "reduced")
    STREND = (8, "s-trend")
    MTREND = (16, "m-trend")
    TEST_POINT = (32, "test-pt")
    STATIC = (64, "static")

NDS2_CHANNEL_TYPE_ANY = Nds2ChannelType.any().value


class Nds2DataType(NumpyTypeEnum, _Nds2Enum):
    """NDS2 data type with descriptive name."""

    @classmethod
    def find(cls, name: type | str | int) -> Self:
        """Return the NDS2 type enumerator corresponding to the given name."""
        try:
            return super().find(name)
        except ValueError:
            # Return unknown
            return cls(0)  # type: ignore[call-arg]

    UNKNOWN = (0, "UNKNOWN")
    INT16 = (1, "int_2")
    INT32 = (2, "int_4")
    INT64 = (4, "int_8")
    FLOAT32 = (8, "real_2")
    FLOAT64 = (16, "real_4")
    COMPLEX32 = (32, "complex_8")
    UINT32 = (64, "uint_4")

NDS2_DATA_TYPE_ANY = Nds2DataType.any().value


# -- warning suppression -------------

class NDSWarning(UserWarning):
    """Warning about communicating with the Network Data Server."""


# always display these warnings
warnings.simplefilter("always", NDSWarning)


# -- query utilities -----------------

def _get_nds2_name(
    channel: str | Channel | nds2.channel,
) -> str:
    """Return the NDS2-formatted name for a channel.

    Understands how to format NDS name strings from
    `gwpy.detector.Channel` and `nds2.channel` objects
    """
    from ..detector import Channel

    if isinstance(channel, Channel):
        return channel.ndsname

    import nds2

    if isinstance(channel, nds2.channel):
        return ",".join((
            channel.name,
            channel.channel_type_to_string(channel.channel_type),
        ))
    return str(channel)


def _get_nds2_names(
    channels: Iterable[str | Channel | nds2.channel],
) -> Iterator[str]:
    """Map `_get_nds2_name` for a list of input channels."""
    return map(_get_nds2_name, channels)


# -- connection utilities ------------

def parse_nds_env(
    env: str = "NDSSERVER",
) -> list[tuple[str, int | None]]:
    """Parse the NDSSERVER environment variable into a list of hosts.

    Parameters
    ----------
    env : `str`, optional
        Shell environment variable name to use for server order.
        Default is ``"NDSSERVER"``.
        The contents of this variable should be a comma-separated
        list of ``host:port`` strings, e.g.
        ``"nds1.server.com:80,nds2.server.com:80"``.

    Returns
    -------
    hostiter : `list` of `tuple`
        a list of (unique) ``(str, int)`` tuples for each host:port
        pair
    """
    hosts = []
    for item in os.environ[env].split(","):
        try:
            host, portstr = item.rsplit(":", 1)
        except ValueError:
            port = None
            host = str(item)
        else:
            port = int(portstr)
        if (host, port) not in hosts:
            hosts.append((host, port))
    return hosts


def host_resolution_order(
    ifo: str | None,
    env: str | None = "NDSSERVER",
    epoch: SupportsToGps = "now",
    lookback: float = 14*86400,
    *,
    include_gwosc: bool = True,
) -> list[tuple[str, int | None]]:
    """Generate a logical ordering of NDS (host, port) tuples for this IFO.

    Parameters
    ----------
    ifo : `str`
        Prefix for IFO of interest.

    env : `str`, optional
        Shell environment variable name to use for server order.
        Default is ``"NDSSERVER"``.
        The contents of this variable should be a comma-separated
        list of ``host:port`` strings, e.g.
        ``"nds1.server.com:80,nds2.server.com:80"``.
        Pass ``env=None`` to disable parsing any environment variable.

    epoch : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS epoch of data requested.

    lookback : `float`, optional
        Duration of spinning-disk cache. This value triggers defaulting to
        the CIT NDS2 server over those at the LIGO sites.
        Default is two weeks.

    include_gwosc : `bool`, optional
        If `True` include the |GWOSC|_ NDS2 instance in the list.

    Returns
    -------
    hro : `list` of `2-tuples <tuple>`
        Drdered `list` of ``(host, port)`` tuples.

    Examples
    --------
    With no environment settings:

    >>> host_resolution_order('H1')
    [('nds.ligo-wa.caltech.edu', None),
     ('nds.ligo.caltech.edu', None),
     ('nds.gwosc.org', None),
    ]
    >>> host_resolution_order('H1', include_gwosc=False))
    [('nds.ligo-wa.caltech.edu', None), ('nds.ligo.caltech.edu', None)]
    """
    hosts = []

    # if given environment variable exists, it will contain a
    # comma-separated list of host:port strings giving the logical ordering
    if env and os.getenv(env):
        hosts = parse_nds_env(env)

    # add our ordered list based on the IFO and the lookback time
    # (at LIGO, the LIGO-Caltech data centre has more data on disk than
    #  the observatory data centres do, so prefer Caltech for older data)
    if to_gps("now") - to_gps(epoch) > lookback:
        ifolist = [None, ifo]
    else:
        ifolist = [ifo, None]

    for difo in ifolist:
        try:
            ifodefault = DEFAULT_HOSTS[difo]
        except KeyError:
            # unknown default NDS2 host for detector, if we don't have
            # hosts already defined (either by NDSSERVER or similar)
            # we should warn the user
            if not hosts:
                warnings.warn(
                    f"no default host found for ifo '{ifo}'",
                    stacklevel=2,
                )
        else:
            if ifodefault not in hosts:
                hosts.append(ifodefault)

    # append the GWOSC NDS2 instance(s) to the list, just in case.
    if include_gwosc:
        hosts.extend(GWOSC_NDS2_HOSTS)

    return hosts


def connect(
    host: str,
    port: int | None = None,
) -> nds2.connection:
    """Open an `nds2.connection` to a given host and port.

    Parameters
    ----------
    host : `str`
        Name of server with which to connect.

    port : `int`, optional
        Connection port.

    Returns
    -------
    connection : `nds2.connection`
        A new open connection to the given NDS host.
    """
    import nds2

    logger.debug("Connecting to NDS server %s", host)
    # set default port for NDS1 connections (required, I think)
    if port is None and NDS1_HOSTNAME.match(host):
        port = 8088

    if port is None:
        conn = nds2.connection(host)
    else:
        conn = nds2.connection(host, port)
    logger.debug(
        "Connected to NDS%d server %s",
        conn.get_protocol(),
        conn.get_host(),
    )
    return conn


def auth_connect(
    host: str,
    port: int | None = None,
) -> nds2.connection:
    """Open an `nds2.connection` handling simple authentication errors.

    This method will catch exceptions related to kerberos authentication,
    and execute a kinit() for the user before attempting to connect again.

    Parameters
    ----------
    host : `str`
        Name of server with which to connect.

    port : `int`, optional
        Connection port.

    Returns
    -------
    connection : `nds2.connection`
        A new open connection to the given NDS host.

    See Also
    --------
    gwpy.io.nds2.connect
        For details of opening connections.
    """
    try:
        return connect(host, port)
    except RuntimeError as exc:
        if "Request SASL authentication" not in str(exc):
            raise
    logger.warning(
        "error authenticating against %s, attempting Kerberos kinit()",
        host,
    )
    kinit()
    return connect(host, port)


@contextmanager
def _connection(
    connection: nds2.connection | None = None,
    host: str | None = None,
    port: int | None = None,
) -> Iterator[nds2.connection]:
    """Yield an `nds2.connection`, opening a new one as required.

    The function is a context manager designed to be used with the `with` statement.

    If a connection is given, it will **not** be closed automatically;
    if one is created by this function, it will be closed.
    """
    if connection is None:
        # Open a new connection
        if host is None:
            msg = "one of `connection` or `host` is required to query NDS2 server"
            raise ValueError(msg)
        conn = auth_connect(host, port)
        epoch = None
    else:
        conn = connection
        epoch = conn.current_epoch()

    yield conn

    if connection is None:
        # Close connections we opened
        conn.close()
    if epoch is not None:
        # Otherwise reset the epoch to what it was before we started
        conn.set_epoch(epoch.gps_start, epoch.gps_stop)


def _parse_nds2_enums(func: Callable[P, T]) -> Callable[P, T]:
    """Decorate a function to translate a type string into an integer."""
    @wraps(func)
    def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> T:
        for kwd, enum_ in (
            ("type", Nds2ChannelType),
            ("dtype", Nds2DataType),
        ):
            val = cast("str | int | None", kwargs.get(kwd))
            if val is None:
                kwargs[kwd] = enum_.any()
            elif not isinstance(val, int):
                kwargs[kwd] = enum_.find(val).value
        return func(*args, **kwargs)
    return wrapped_func


# -- query methods -------------------

@_parse_nds2_enums
def find_channels(
    channels: Sequence[str | Channel | nds2.channel],
    connection: nds2.connection | None = None,
    host: str | None = None,
    port: int | None = None,
    sample_rate: float | tuple[float, float] | None = None,
    type: int | str | None = NDS2_CHANNEL_TYPE_ANY,  # noqa: A002
    dtype: int | str | DTypeLike = NDS2_DATA_TYPE_ANY,
    *,
    unique: bool = False,
    epoch: str | tuple[int, int] | None = "ALL",
) -> list[nds2.channel]:
    """Query an NDS2 server for channel information.

    Parameters
    ----------
    channels : `list` of `str`
        List of channel names to query, each can include bash-style globs.

    connection : `nds2.connection`, optional
        Open NDS2 connection to use for query.

    host : `str`, optional
        Name of NDS2 server to query, required if ``connection`` is not
        given.

    port : `int`, optional
        Port number on host to use for NDS2 connection.

    sample_rate : `int`, `float`, `tuple`, optional
        A single number, representing a specific sample rate to match,
        or a tuple representing a ``(low, high)`` interval to match.

    type : `int`, optional
        The NDS2 channel type to match.

    dtype : `int`, optional
        The NDS2 data type to match.

    unique : `bool`, optional, default: `False`
        Require one (and only one) match per channel.

    epoch : `str`, `tuple` of `int`, optional
        The NDS epoch to restrict to, either the name of a known epoch,
        or a 2-tuple of GPS ``[start, stop)`` times.

    Returns
    -------
    channels : `list` of `nds2.channel`
        List of NDS2 channel objects.

    See Also
    --------
    nds2.connection.find_channels
        For documentation on the underlying query method.

    Examples
    --------
    >>> from gwpy.io.nds2 import find_channels
    >>> find_channels(['G1:DER_DATA_H'], host='nds.ligo.caltech.edu')
    [<G1:DER_DATA_H (16384Hz, RDS, FLOAT64)>]
    """
    # by the time we get here, the decorator will have converted
    # type and dtype to integers
    type = cast("int", type)  # noqa: A001
    dtype = cast("int", dtype)

    with _connection(connection, host, port) as conn:
        # set epoch
        if isinstance(epoch, tuple):
            conn.set_epoch(*epoch)
        else:
            conn.set_epoch(epoch or "ALL")

        # format {min,max}_sample_rate options
        kwargs = {}
        if isinstance(sample_rate, int | float):
            kwargs["min_sample_rate"] = sample_rate
            kwargs["max_sample_rate"] = sample_rate
        elif sample_rate is not None:
            kwargs["min_sample_rate"], kwargs["max_sample_rate"] = sample_rate

        # query for channels
        out = []
        for name in _get_nds2_names(channels):
            out.extend(_find_channel(
                conn,
                channel_glob=name,
                channel_type_mask=type,
                data_type_mask=dtype,
                unique=unique,
                **kwargs,
            ))
        return out


def _find_channel(
    connection: nds2.connection,
    channel_glob: str = "*",
    channel_type_mask: int = Nds2ChannelType.any().value,
    data_type_mask: int = Nds2DataType.any().value,
    min_sample_rate: float = MIN_SAMPLE_RATE,
    max_sample_rate: float = MAX_SAMPLE_RATE,
    *,
    unique: bool = False,
) -> list[nds2.channel]:
    """Find a single channel.

    Parameters
    ----------
    connection : `nds2.connection`, optional
        Open NDS2 connection to use for query

    channel_glob : `str`, optional
        The name of the channel to find.

    channel_type_mask : `int`, optional
        The NDS2 channel type to match.

    data_type_mask : `int`, optional
        The NDS2 data type to match.

    min_sample_rate : `float`, optional
        The lowest sample rate to match.

    max_sample_rate : `float`, optional
        The highest sample rate to match.

    unique : `bool`, optional
        Require one (and only one) match per channel.

    Returns
    -------
    channels : `list` of `nds2.channel`
        List of NDS2 channel objects, if `unique=True` is given the list
        is guaranteed to have only one element.

    Raises
    ------
    ValueError
        If ``unique=True`` and more than one channel is found.

    See Also
    --------
    nds2.connection.find_channels
        For documentation of all arguments and on the underlying query method.
    """
    # parse channel type from name,
    # e.g. 'L1:GDS-CALIB_STRAIN,reduced' -> 'L1:GDS-CALIB_STRAIN', 'reduced'
    channel_glob, channel_type_mask = _strip_ctype(
        channel_glob,
        channel_type_mask,
        connection.get_protocol(),
    )

    # query NDS2
    found = connection.find_channels(
        channel_glob,
        channel_type_mask,
        data_type_mask,
        min_sample_rate,
        max_sample_rate,
    )

    # if don't care about defaults, just return now
    if not unique:
        return found

    # if two results, remove 'online' copy (if present)
    #    (if no online channels present, this does nothing)
    if len(found) == 2:  # noqa: PLR2004
        online = Nds2ChannelType.ONLINE.value  # type: ignore[attr-defined]
        found = [c for c in found if c.channel_type != online]

    # if not unique result, panic
    if len(found) != 1:
        msg = f"unique NDS2 channel match not found for '{channel_glob}'"
        raise ValueError(msg)

    return found


def _strip_ctype(
    name: str,
    ctype: int,
    protocol: int = 2,
) -> tuple[str, int]:
    """Strip the ctype from a channel name for the given nds server version.

    This is needed because NDS1 servers store trend channels _including_
    the suffix, but not raw channels, and NDS2 doesn't do this.
    """
    # parse channel type from name (e.g. 'L1:GDS-CALIB_STRAIN,reduced')
    try:
        name, ctypestr = name.rsplit(",", 1)
    except ValueError:
        return name, ctype

    ctype = Nds2ChannelType.find(ctypestr).value

    # NDS1 stores channels with trend suffix, so we put it back:
    if protocol == 1 and ctype in (
        Nds2ChannelType.STREND.value,  # type: ignore[attr-defined]
        Nds2ChannelType.MTREND.value,  # type: ignore[attr-defined]
    ):
        name += f",{ctypestr}"

    return name, ctype


def get_availability(
    channels: Sequence[str | Channel | nds2.channel],
    start: int,
    end: int,
    connection: nds2.connection | None = None,
    host: str | None = None,
    port: int | None = None,
) -> SegmentListDict:
    """Query an NDS2 server for data availability.

    Parameters
    ----------
    channels : `list` of `str`
        List of channel names to query; this list is mapped to NDS channel
        names using :func:`find_channels`..

    start : `int`
        GPS start time of query.

    end : `int`
        GPS end time of query.

    connection : `nds2.connection`, optional
        Open NDS2 connection to use for query.

    host : `str`, optional
        Name of NDS2 server to query, required if ``connection`` is not
        given.

    port : `int`, optional
        Port number on host to use for NDS2 connection.

    Returns
    -------
    segdict : `~gwpy.segments.SegmentListDict`
        Dict of ``(name, SegmentList)`` pairs.

    Raises
    ------
    ValueError
        If the given channel name cannot be mapped uniquely to a name
        in the NDS server database.

    See Also
    --------
    nds2.connection.get_availability
        For documentation on the underlying query method.
    """
    from ..segments import (
        Segment,
        SegmentList,
        SegmentListDict,
    )

    with _connection(connection, host, port) as conn:
        conn.set_epoch(start, end)

        # map user-given real names to NDS names
        names = list(map(_get_nds2_name, find_channels(
            channels,
            epoch=(start, end),
            connection=conn,
            unique=True,
        )))

        # query for availability
        result = conn.get_availability(names)

    # map to segment types
    out = SegmentListDict()
    for name, res in zip(channels, result, strict=True):
        out[name] = SegmentList([
            Segment(s.gps_start, s.gps_stop)
            for s in res.data
        ])
    return out


def minute_trend_times(start: int, end: int) -> tuple[int, int]:
    """Expand a [start, end) interval for use in querying for minute trends.

    NDS2 requires start and end times for minute trends to be a multiple of
    60 (to exactly match the time of a minute-trend sample), so this function
    expands the given ``[start, end)`` interval to the nearest multiples.

    Parameters
    ----------
    start : `int`
        GPS start time of query.

    end : `int`
        GPS end time of query.

    Returns
    -------
    mstart : `int`
        ``start`` rounded down to nearest multiple of 60.
    mend : `int`
        ``end`` rounded up to nearest multiple of 60.

    Examples
    --------
    >>> minute_trend_times(123, 456)
    (120, 480)
    """
    if start % 60:
        start = int(start) // 60 * 60
    if end % 60:
        end = int(end) // 60 * 60 + 60
    return int(start), int(end)



# Copyright (c) 2022-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWDataFind UI for FFL cache files.

This module is used to replace the proper GWDataFind interface
on-the-fly when FFL data access is inferred.
As such this module is required to emulate those functions
from `gwdatafind` used in :mod:`gwpy.io.datafind`.
"""

from __future__ import annotations

import logging
import os
import re
from collections import defaultdict
from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING
from warnings import warn

from igwn_segments import (
    segment,
    segmentlist,
)

from .cache import (
    _CacheEntry,
    _iter_cache,
    cache_segments,
    file_segment,
    read_cache_entry,
)

if TYPE_CHECKING:
    from collections.abc import Generator

    from .utils import FileSystemPath

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

logger = logging.getLogger(__name__)

_SITE_REGEX = re.compile(r"\A(\w+)-")
_DEFAULT_TYPE_MATCH = re.compile(r"^(?!lastfile|spectro|\.).*")


# -- generic utilities ---------------

@cache
def _read_last_line(
    path: FileSystemPath,
    bufsize: int = 2,
    encoding: str = "utf-8",
) -> str:
    """Read and return the last line of a text file."""
    with open(path, "rb") as fobj:  # noqa: PTH123
        # start from the end and read until we find a full line, or reach the beginning
        fobj.seek(0, os.SEEK_END)
        count = 0
        try:
            while fobj.read(1) != b"\n":
                fobj.seek(-bufsize, os.SEEK_CUR)
                count += 1
        except OSError:
            if not count:  # file is empty or unreadable
                raise
            # if we've rewound to the start of the file, just stop
            if fobj.tell() < bufsize:
                fobj.seek(0)
            # otherwise this is a different error
            else:
                raise
        return fobj.read().decode(encoding).strip()


# -- ffl utilities -------------------

def _get_ffl_basedir() -> Path:
    """Return the base directory containing FFL cache files."""
    # FFLPATH
    try:
        fflpath = Path(os.environ["FFLPATH"])
    except KeyError:
        pass
    else:
        logger.debug("Parsed FFLPATH='%s'", fflpath)
        return fflpath

    # VIRGODATA
    try:
        virgodata = Path(os.environ["VIRGODATA"])
    except KeyError:
        msg = (
            "failed to parse FFLPATH from environment, please set "
            "FFLPATH to point to the directory containing FFL files"
        )
        raise KeyError(msg) from None
    else:
        logger.debug("Parsed VIRGODATA='%s'", virgodata)
        return virgodata / "ffl"


def _is_ffl_file(path: FileSystemPath) -> bool:
    """Return True if the given path looks like an FFL file path."""
    return Path(path).suffix == ".ffl"


def _get_site_tag(path: FileSystemPath) -> tuple[str, str]:
    """Return the (site, tag) tuple given the path to an FFL file."""
    path = Path(path)
    tag = path.stem
    last = _read_last_line(path).split()[0]
    match = _SITE_REGEX.match(Path(last).name)
    if match is None:
        msg = f"Site tag not found in last line of FFL file: {last}"
        raise ValueError(msg)
    site = match.groups()[0]
    return site, tag


def _find_ffl_files(
    basedir: FileSystemPath | None = None,
) -> Generator[str, None, None]:
    """Yield all FFL file paths from the given base directory."""
    if basedir is None:
        basedir = _get_ffl_basedir()
    basedir = Path(basedir)
    for root, _, names in os.walk(basedir):
        for name in filter(_is_ffl_file, names):
            yield str(Path(root) / name)


@cache
def _find_ffls(basedir: str | Path | None = None) -> dict[tuple[str, str], list[str]]:
    """Find all readable FFL files."""
    ffls = defaultdict(list)
    for path in _find_ffl_files(basedir=basedir):
        try:
            site, tag = _get_site_tag(path)
        except (
            OSError,  # file is empty (or cannot be read at all)
            ValueError,  # last entry didn't match _SITE_REGEX
        ):
            continue
        logger.debug("Found FFL for %s-%s: '%s'", site, tag, path)
        ffls[(site, tag)].append(path)
    return ffls


def _ffl_paths(
    site: str,
    tag: str,
    basedir: str | Path | None = None,
) -> list[str]:
    """Return the paths of all FFL files for a given site and tag."""
    try:
        return _find_ffls(basedir=basedir)[(site, tag)]
    except KeyError as exc:
        msg = f"no FFL file found for ('{site}', '{tag}')"
        raise ValueError(msg) from exc


@cache
def _read_ffls(
    site: str,
    tag: str,
    basedir: str | Path | None = None,
) -> list[_CacheEntry]:
    """Read all FFL files for a given site and tag as a list of CacheEntry objects.

    Parameters
    ----------
    site : `str`
        Observatory ID to search for.
    tag : `str`
        Data type tag to search for.
    basedir : `str`, `Path`, or `None`, optional
        Base directory to search in. If `None`, uses default FFL directory.

    Returns
    -------
    entries : `list` of `_CacheEntry`
        List of cache entries found in the FFL files.
    """
    entries: list[_CacheEntry] = []
    for ffl in _ffl_paths(site, tag, basedir=basedir):
        with Path(ffl).open() as fobj:
            entries.extend(
                _CacheEntry(site, tag, entry.segment, entry.path)
                for entry in _iter_cache(fobj, gpstype=float)
            )
    return entries


def _handle_error(action: str, message: str) -> None:
    """Handle error, warn, or ignore for the given state.

    Parameters
    ----------
    action : `str`
        The action to perform, one of
        ``'warn'`` (emit a `UserWarning`, default),
        ``'ignore'`` (do nothing),
        or anything else (raise a `RuntimeError`).

    message : `str`
        The message to emit with warnings or errors.

    Raises
    ------
    RuntimeError
        If action is not ``'warn'`` or ``'ignore'``.
    """
    # if ignore, do nothing
    if action == "ignore":
        return
    # if warn, emit a warning
    if action == "warn":
        warn(message, stacklevel=2)
        return

    # otherwise, raise an error
    msg = message
    raise RuntimeError(msg)


# -- ui ---------------------

def _check_gwdatafind_kwargs(
    urltype: str,
    ext: str | None,
) -> None:
    """Validate the arguments that are supported for compatibility with GWDataFind."""
    if urltype != "file":
        msg = "Data discovery via FFL only supports urltype='file'"
        raise ValueError(msg)
    if ext not in {"gwf", None}:
        msg = "Data discovery via FFL only supports ext='gwf'"
        raise ValueError(msg)


def find_types(
    site: str | None = None,
    match: str | re.Pattern | None = _DEFAULT_TYPE_MATCH,
    ext: str | None = "gwf",
) -> list[str]:
    """Return the list of known data types.

    Parameters
    ----------
    site : `str`, optional
        Observatory ID (e.g. ``'A'``) to restrict types, if `None` (default)
        is given, all types are returned.

    match : `str`, optional
        Regular expression to use to restrict types.

    ext : `str`, optional
        The file extension for which to search.

    Returns
    -------
    types : `list` of `str`
        The list of data types matching the criteria.
    """
    _check_gwdatafind_kwargs("file", ext)
    ffls = _find_ffls()
    types = [tag for (site_, tag) in ffls if site in (None, site_)]
    if match is not None:
        match = re.compile(match)
        return list(filter(match.search, types))
    return types


def find_urls(
    site: str,
    tag: str,
    gpsstart: int,
    gpsend: int,
    match: str | re.Pattern | None = None,
    on_gaps: str = "warn",
    urltype: str = "file",
    ext: str | None = "gwf",
) -> list[str]:
    """Return the list of all files for ``tag`` in the ``[start, end)`` GPS interval.

    Parameters
    ----------
    site : `str`
        Observatory ID to search for.

    tag : `str`
        Data type tag to search for.

    gpsstart : `int`
        GPS start time of query.

    gpsend : `int`
        GPS end time of query.

    match : `str`, optional
        Regular expression to use to restrict returned data URLs.

    on_gaps : `str`, optional
        Action to take when the requested all or some of the GPS interval
        is not covereed by the dataset, one of:

        - ``'error'``: raise a `RuntimeError` (default)
        - ``'warn'``: print a warning but return all available URLs
        - ``'ignore'``: return the list of URLs with no warnings

    urltype : `str`, optional
        Type of URL for which to search. Only ``"file"`` is acceptable.

    ext : `str`, optional
        The file extension for which to search.

    Returns
    -------
    urls : `list` of `str`
        A list of URLs representing discovered data.
    """
    _check_gwdatafind_kwargs(urltype, ext)

    if match:
        match = re.compile(match)

    span = segment(gpsstart, gpsend)

    def _keep(e: _CacheEntry) -> bool:
        """Return `True` if this `_CacheEntry` is to be kept."""
        return bool(
            e.observatory == site
            and e.path.endswith(f".{ext}")
            and e.description == tag
            and e.segment.intersects(span)
            and (match.search(e.path) if match else True),
        )

    cache = list(filter(_keep, _read_ffls(site, tag)))
    urls = [e.path for e in cache]

    # handle missing data
    missing = segmentlist([span]) - cache_segments(cache)
    if missing:
        msg = "Missing segments: \n" + "\n".join(map(str, missing))
        _handle_error(on_gaps, msg)

    return urls


def find_latest(
    site: str,
    tag: str,
    urltype: str = "file",
    ext: str | None = "gwf",
    on_missing: str = "warn",
) -> list[str]:
    """Return the most recent file of a given type.

    Parameters
    ----------
    site : `str`
        Observatory ID to search for.

    tag : `str`
        Data type tag to search for.

    urltype : `str`, optional
        Type of URL for which to search. Only ``"file"`` is acceptable.

    ext : `str`, optional
        The file extension for which to search.

    on_missing : `str`, optional
        What to do if a URL is not found for the given site and tag, one of
        ``'warn'`` (emit a `UserWarning`, default),
        ``'ignore'`` (do nothing),
        anything else (raise a `RuntimeError`).

    Returns
    -------
    urls : `list` of `str`
        A list (typically of one item) of URLs representing the latest data
        for a specific site and tag.
    """
    _check_gwdatafind_kwargs(urltype, ext)

    try:
        fflfiles = _ffl_paths(site, tag)
    except ValueError:  # no readable FFL file
        urls = []
    else:
        urls = [
            read_cache_entry(_read_last_line(fflfile), gpstype=float)
            for fflfile in fflfiles
        ]
        if urls:  # if multiple, find the latest one
            urls = sorted(urls, key=file_segment)[-1:]

    if not urls:
        _handle_error(on_missing, "No files found")

    return urls



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Test setup for gwpy."""

from __future__ import annotations

import contextlib
import warnings
from typing import TYPE_CHECKING

from matplotlib import (
    rcParams,
    use,
)

if TYPE_CHECKING:
    import pytest

# force Agg for all tests
use("agg", force=True)

# register custom fixtures for all test modules
from .testing.fixtures import *  # noqa: F403

# define marks (registered below)
from .testing.marks import _register_marks

# -- plotting options

# ignore errors due from pyplot.show() using Agg
warnings.filterwarnings("ignore", message=".*non-GUI backend.*")

# force simpler rcParams for all tests
# (fixtures or tests may update these individually)
# NOTE: this most-likely happens _after_ gwpy.plot has
#       updated the rcParams once, so these settings should persist
rcParams.update({
    "text.usetex": False,  # TeX is slow most of the time
})

# -- gpstime compatibility
# import gpstime now so that the first import is never during a
# pytest run using xdist where the multiple workers try and download
# the leap seconds file at the same time, and then fall over each other

with contextlib.suppress(ImportError):
    import gpstime  # noqa: F401


# -- pytest configuration

def pytest_configure(config: pytest.Config) -> None:
    """Configure pytest with custom marks."""
    _register_marks(config)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from the PyCBC live online GW search."""

from __future__ import annotations

import re
from pathlib import Path
from typing import TYPE_CHECKING

import h5py
import numpy
from astropy.table import Table

from ...io.hdf5 import (
    identify_hdf5,
    with_read_hdf5,
)
from .. import EventTable
from ..filter import (
    filter_table,
    parse_column_filters,
)
from .utils import (
    DYNAMIC_COLUMN_FUNC,
    DYNAMIC_COLUMN_INPUT,
)

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import (
        Any,
        Literal,
    )

    from ...frequencyseries import FrequencySeries
    from ...io.utils import (
        FileLike,
        FileSystemPath,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "Alex Nitz <alex.nitz@ligo.org>"

PYCBC_LIVE_FORMAT = "hdf5.pycbc_live"

META_COLUMNS: set[str] = {
    "loudest",
    "psd",
}
PYCBC_FILENAME = re.compile(
    "([A-Z][0-9])+-Live-[0-9.]+-[0-9.]+.(h5|hdf|hdf5)",
)


def _get_column(
    source: h5py.Group,
    name: str,
    loudest: numpy.ndarray | None = None,
) -> Table.Column:
    """Read a `~astropy.table.Column` from a PyCBC `h5py.Group`."""
    try:  # normal column
        arr = source[name][:]
    except KeyError:
        # try and generate the column on-the-fly
        if name in GET_COLUMN:
            arr = GET_COLUMN[name](source)
        else:
            raise
    if loudest is not None:  # return only the loudest events
        arr = arr[loudest]
    return Table.Column(arr, name=name)


@with_read_hdf5
def table_from_file(
    source: h5py.Group,
    ifo: str | None = None,
    columns: list[str] | None = None,
    *,
    where: str | list[str] | None = None,
    loudest: bool = False,
    extended_metadata: bool = True,
) -> Table:
    """Read a `Table` from a PyCBC live HDF5 file.

    The PyCBC project stores tabular data in HDF5 files using an `h5py.Group`
    to describe the table, and `h5py.Dataset` structures for each column.

    Parameters
    ----------
    source : `str`, `pathlib.Path`, `file`, `h5py.Group`
        The file path or open `h5py` object from which to read the data.

    ifo : `str`, optional
        The interferometer prefix (e.g. ``'G1'``) to read; this is required
        if reading from a file path or a `h5py.File` and the containing file
        stores data for multiple interferometers.

    columns : `list` of `str`, optional
        The list of column names to include in returned table.
        Default is all columns.

    where : `str`, `list` of `str`, optional
        One or more filter condition strings to apply, e.g. ``'snr>6'``.

    loudest : `bool`, optional
        If `True` read only those events marked as 'loudest'.
        Default is `False` (read all rows in the table).

    extended_metadata : `bool`, optional
        If `True` record non-column datasets found in the H5 group (e.g. ``'psd'``)
        in the ``meta`` dict. Default is `True`.

    Returns
    -------
    table : `~astropy.table.Table`
    """
    # find group
    if isinstance(source, h5py.File):
        source, ifo = _find_table_group(source, ifo=ifo)

    # -- by this point 'source' is guaranteed to be an h5py.Group

    # list of columns to read (default: everything that's there)
    read_cols = set(columns or _get_columns(source))

    # parse where conditions
    filters = parse_column_filters(where or [])
    for fs in filters:
        cols = fs.column
        if isinstance(cols, str):
            read_cols.add(cols)
        else:
            read_cols.update(cols)

    # set up meta dict
    meta = {"ifo": ifo}
    meta.update(source.attrs)
    if extended_metadata:
        meta.update(_get_extended_metadata(source))

    # extract loudest array index
    loudidx: numpy.ndarray | None = None
    if loudest:
        loudidx = source["loudest"][:]

    # map data to columns
    data = [
        _get_column(source, name, loudest=loudidx)
        for name in read_cols
    ]

    # read, applying filters, and column filters
    tab = Table(data, meta=meta)
    if filters:
        tab = filter_table(tab, filters)
    if columns:  # return only the columns the user asked for
        return tab[columns]
    return tab


def _find_table_group(
    h5file: h5py.Group,
    ifo: str | None = None,
) -> tuple[h5py.Group, str]:
    """Find the right `h5py.Group` within the given `h5py.File`."""
    exclude = ("background",)
    if ifo is None:
        try:
            ifo, = (key for key in h5file if key not in exclude)
        except ValueError as exc:
            exc.args = (
                "PyCBC live HDF5 file contains dataset groups "
                "for multiple interferometers, please specify "
                "the prefix of the relevant interferometer via "
                "the `ifo` keyword argument, e.g: `ifo=G1`",
            )
            raise
    try:
        return h5file[ifo], ifo
    except KeyError as exc:
        exc.args = (f"No group for ifo '{ifo}' in PyCBC live HDF5 file",)
        raise


def _get_columns(h5group: h5py.Group) -> set[str]:
    """Find valid column names from a PyCBC HDF5 Group.

    Returns a `set` of names.
    """
    columns = set()
    for name in sorted(h5group):
        # not a column, or one to ignore
        if (
            not isinstance(h5group[name], h5py.Dataset)
            or name == "template_boundaries"
        ):
            continue

        # template parameters, ignore those
        if (
            name.endswith("_template")
            and name[:-9] in columns
        ):
            continue

        columns.add(name)
    return columns - META_COLUMNS


def _parse_psd(dataset: h5py.Dataset) -> FrequencySeries:
    """Parse a PyCBC PSD from an `h5py.Dataset`.

    Returns
    -------
    gwpy.frequencyseries.FrequencySeries
    """
    from gwpy.frequencyseries import FrequencySeries
    return FrequencySeries(
        dataset[:],
        f0=0,
        df=dataset.attrs["delta_f"],
        name="pycbc_live",
    )


def _get_extended_metadata(h5group: h5py.Group) -> dict[str, Any]:
    """Extract the extended metadata for a PyCBC table in HDF5.

    This method packs non-table-column datasets in the given h5group into
    a metadata `dict`.

    Returns
    -------
    meta : `dict`
       The metadata dict.
    """
    meta = {}

    # get PSD
    try:
        psdds = h5group["psd"]
    except KeyError:
        pass
    else:
        meta["psd"] = _parse_psd(psdds)

    # get everything else
    for key in META_COLUMNS - {"psd"}:
        try:
            value = h5group[key][:]
        except KeyError:
            pass
        else:
            meta[key] = value

    return meta


def filter_empty_files(
    files: list[str],
    ifo: str | None = None,
) -> list[str]:
    """Remove empty PyCBC-HDF5 files from a list.

    Parameters
    ----------
    files : `list`
        A list of file paths to test.

    ifo : `str`, optional
        Prefix for the interferometer of interest (e.g. ``'L1'``),
        include this for a more robust test of 'emptiness'.

    Returns
    -------
    nonempty : `list`
        The subset of the input ``files`` that are considered not empty.

    See Also
    --------
    empty_hdf5_file
        For details of the 'emptiness' test.
    """
    return type(files)([f for f in files if not empty_hdf5_file(f, ifo=ifo)])


@with_read_hdf5
def empty_hdf5_file(
    h5f: h5py.File,
    ifo: str | None = None,
) -> bool:
    """Determine whether a PyCBC-HDF5 file is empty.

    A file is considered empty if it contains no groups at the base level,
    or if the ``ifo`` group contains only the ``psd`` dataset.

    Parameters
    ----------
    h5f : `str`, `pathlib.Path`, `h5py.File`
        The path of the pycbc_live file to test.

    ifo : `str`, optional
        Prefix for the interferometer of interest (e.g. ``'L1'``),
        include this for a more robust test of 'emptiness'.

    Returns
    -------
    empty : `bool`
        `True` if the file looks to have no content, otherwise `False`.
    """
    # the decorator opens the HDF5 file for us, so h5f is guaranteed to
    # be an h5py.Group object
    h5f = h5f.file
    # if the root group is empty, the file is empty
    if list(h5f) == []:
        return True

    # for each group (or the IFO group given by the user),
    # check whether there is any useful content
    groups = h5f.keys() if ifo is None else [ifo]
    return all(set(h5f.get(group, [])) <= {"gates", "psd"} for group in groups)


def identify_pycbc_live(
    origin: Literal["read", "write"],
    filepath: FileSystemPath | None,
    fileobj: FileLike | None,
    *args,  # noqa: ANN002
    **kwargs,
) -> bool:
    """Identify a PyCBC Live file as an HDF5 with the correct name."""
    # first, check that this is a valid HDF5 file
    if not identify_hdf5(origin, filepath, fileobj, *args, **kwargs):
        return False

    # next, check that the filename matches what we expect from PyCBC live
    # -- this is terrible, but it's the best I can think of
    if filepath is None and fileobj is not None:  # need file name
        filepath = getattr(fileobj, "name", None)
    return bool(
        filepath is not None
        and PYCBC_FILENAME.match(Path(filepath).name),
    )


# register for unified I/O (with higher priority than HDF5 reader)
EventTable.read.registry.register_identifier(
    PYCBC_LIVE_FORMAT,
    EventTable,
    identify_pycbc_live,
)
EventTable.read.registry.register_reader(
    PYCBC_LIVE_FORMAT,
    EventTable,
    table_from_file,
    priority=1,
)

# -- processed columns ---------------
#
# Here we define methods required to build commonly desired columns that
# are just a combination of the basic columns.
#
# Each method should take in an `~h5py.Group` and return a `numpy.ndarray`

GET_COLUMN: dict[str, Callable] = {}
GET_COLUMN_EXTRA: dict[str, set[str]] = {}


def _new_snr_scale(
    redx2: float | numpy.ndarray,
    q: float = 6.,
    n: float = 2.,
) -> float | numpy.ndarray:
    return (.5 * (1. + redx2 ** (q/n))) ** (-1./q)


def get_new_snr(
    h5group: h5py.Group,
    q: float = 6.,
    n: float = 2.,
) -> numpy.ndarray:
    """Calculate the 'new SNR' column for this PyCBC HDF5 table group."""
    newsnr = h5group["snr"][:].copy()
    rchisq = h5group["chisq"][:]
    idx = numpy.where(rchisq > 1.)[0]
    newsnr[idx] *= _new_snr_scale(rchisq[idx], q=q, n=n)
    return newsnr


GET_COLUMN["new_snr"] = get_new_snr
GET_COLUMN_EXTRA["new_snr"] = {"snr", "chisq"}

# use the generic mass functions
for _key in (
    "mchirp",
    "mtotal",
):
    GET_COLUMN[_key] = DYNAMIC_COLUMN_FUNC[_key]
    GET_COLUMN_EXTRA[_key] = DYNAMIC_COLUMN_INPUT[_key]

# update custom columns using pycbc's ranking function dict
try:
    from pycbc.events.ranking import (
        required_datasets,
        sngls_ranking_function_dict,
    )
except ImportError:
    pass
else:
    GET_COLUMN.update(sngls_ranking_function_dict)
    GET_COLUMN_EXTRA.update(required_datasets)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWpy API to the scipy.signal FFT routines."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import numpy
import scipy.signal
from packaging.version import Version
from scipy import __version__ as scipy_version

from ...frequencyseries import FrequencySeries
from . import _registry as fft_registry
from ._utils import scale_timeseries_unit

if TYPE_CHECKING:
    from collections.abc import Callable

    from numpy.typing import NDArray

    from ...timeseries import TimeSeries
    from ..window import WindowLike

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- density scaling methods ---------

def _spectral_density(
    timeseries: TimeSeries,
    segmentlength: int,
    noverlap: int | None = None,
    name: str | None = None,
    sdfunc: Callable = scipy.signal.welch,
    **kwargs,
) -> FrequencySeries:
    """Calculate a generic spectral density of this `TimeSeries`."""
    if (
        Version(scipy_version) >= Version("1.16.0a0")
        and (other := kwargs.get("y")) is not None
        and timeseries.size != other.size
    ):
        # manually zero-pad the shorter array, see
        # https://github.com/scipy/scipy/issues/23036
        if (a := timeseries.size) < (b := other.size):
            timeseries = timeseries.pad((0, b - a), mode="constant")
        else:
            kwargs["y"] = numpy.pad(other, (0, a - b), mode="constant")

    # compute spectral density
    freqs, psd_ = sdfunc(
        timeseries.value,
        fs=timeseries.sample_rate.decompose().value,
        nperseg=segmentlength,
        noverlap=noverlap,
        **kwargs,
    )
    # generate FrequencySeries and return
    unit = scale_timeseries_unit(
        timeseries.unit,
        kwargs.get("scaling", "density"),
    )
    return FrequencySeries(
        psd_,
        unit=unit,
        frequencies=freqs,
        name=(name or timeseries.name),
        epoch=timeseries.epoch,
        channel=timeseries.channel,
    )


def welch(
    timeseries: TimeSeries,
    segmentlength: int,
    average: str = "mean",
    **kwargs,
) -> FrequencySeries:
    """Calculate a PSD using Welch's method."""
    return _spectral_density(
        timeseries,
        segmentlength,
        average=average,
        **kwargs,
    )


def bartlett(
    timeseries: TimeSeries,
    segmentlength: int,
    noverlap: int = 0,
    **kwargs,
) -> FrequencySeries:
    """Calculate a PSD using Bartlett's method."""
    return _spectral_density(
        timeseries,
        segmentlength,
        noverlap=noverlap,
        **kwargs,
    )


def median(
    timeseries: TimeSeries,
    segmentlength: int,
    average: str = "median",
    **kwargs,
) -> TimeSeries:
    """Calculate a PSD using Welch's method with a median average."""
    return _spectral_density(
        timeseries,
        segmentlength,
        average=average,
        **kwargs,
    )


# register
for func in (
    welch,
    bartlett,
    median,
):
    fft_registry.register_method(func, name=func.__name__)
    # DEPRECATED:
    fft_registry.register_method(func, name=f"scipy-{func.__name__}")


# -- others --------------------------

def rayleigh(
    timeseries: TimeSeries,
    segmentlength: int,
    noverlap: int = 0,
    window: WindowLike = "hann",
):
    """Calculate a Rayleigh statistic spectrum.

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        Input `TimeSeries` data.

    segmentlength : `int`
        Number of samples in single average.

    noverlap : `int`
        Number of samples to overlap between segments, passing `None` will
        choose based on the window method, default: ``0``.

    window : `str`, `numpy.ndarray`, optional
        Window function to apply to ``timeseries`` prior to FFT,
        see :func:`scipy.signal.get_window` for details on acceptable
        formats.

    Returns
    -------
    spectrum : `~gwpy.frequencyseries.FrequencySeries`
        Average power `FrequencySeries`.
    """
    stepsize = segmentlength - noverlap
    if noverlap:
        numsegs = 1 + int(
            (timeseries.size - segmentlength)
            / float(noverlap),
        )
    else:
        numsegs = int(timeseries.size // segmentlength)
    tmpdata: NDArray[numpy.float64] = numpy.ndarray(
        (numsegs, int(segmentlength // 2 + 1)),
    )
    for i in range(numsegs):
        tmpdata[i, :] = welch(
            timeseries[i * stepsize:i * stepsize + segmentlength],
            segmentlength,
            window=window,
        )
    std = tmpdata.std(axis=0)
    mean = tmpdata.mean(axis=0)
    return FrequencySeries(
        std / mean,
        unit="",
        copy=False,
        f0=0,
        epoch=timeseries.epoch,
        df=timeseries.sample_rate.value / segmentlength,
        channel=timeseries.channel,
        name=f"Rayleigh spectrum of {timeseries.name}",
    )


def csd(
    timeseries: TimeSeries,
    other: TimeSeries,
    segmentlength: int,
    noverlap: int | None = None,
    **kwargs,
) -> TimeSeries:
    """Calculate the CSD of two `TimeSeries` using Welch's method.

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        Time-series of data.

    other : `~gwpy.timeseries.TimeSeries`
        Time-series of data.

    segmentlength : `int`
        Number of samples in single average.

    noverlap : `int`
        Number of samples to overlap between segments, defaults to 50%.

    kwargs
        Other keyword arguments are passed to :meth:`scipy.signal.csd`

    Returns
    -------
    spectrum : `~gwpy.frequencyseries.FrequencySeries`
        Average power `FrequencySeries`.

    See Also
    --------
    scipy.signal.csd
    """
    return _spectral_density(
        timeseries,
        segmentlength,
        y=other.value,
        noverlap=noverlap,
        name=f"{timeseries.name}---{other.name}",
        sdfunc=scipy.signal.csd,
        **kwargs,
    )


def coherence(
    timeseries: TimeSeries,
    other: TimeSeries,
    segmentlength: int,
    downsample: bool | None = None,
    noverlap: int | None = None,
    **kwargs,
) -> FrequencySeries:
    """Calculate the coherence between two `TimeSeries` using Welch's method.

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        Time-series of data.

    other : `~gwpy.timeseries.TimeSeries`
        Time-series of data.

    segmentlength : `int`
        Number of samples in single average..

    noverlap : `int`
        Number of samples to overlap between segments, defaults to 50%.

    downsample : `bool`
        Downsample the series with higher sampling frequency? SciPy assumes
        that both TimeSeries have the same rate.

    kwargs
        Other keyword arguments are passed to :meth:`scipy.signal.coherence`.

    Returns
    -------
    spectrum : `~gwpy.frequencyseries.FrequencySeries`
        Average power `FrequencySeries`.

    See Also
    --------
    scipy.signal.coherence
    """
    # Should we warn about unequal sampling frequencies?
    warn_fs = False

    if downsample is None:
        warn_fs = True
        downsample = True

    if timeseries.sample_rate != other.sample_rate:
        # scipy assumes a single sampling frequency
        if not downsample:
            msg = (
                "Cannot calculate coherence when sampling "
                "frequencies are unequal"
            )
            raise ValueError(msg)
        if warn_fs:
            warnings.warn(
                "Sampling frequencies are unequal. Higher "
                "frequency series will be downsampled before "
                "coherence is calculated",
                category=UserWarning,
                stacklevel=2,
            )
        # downsample the one with the higher rate
        if timeseries.sample_rate > other.sample_rate:
            timeseries = timeseries.resample(other.sample_rate)
        else:
            other = other.resample(timeseries.sample_rate)

    # calculate CSD
    out = _spectral_density(
        timeseries,
        segmentlength,
        y=other.value,
        noverlap=noverlap,
        name=f"Coherence between {timeseries.name} and {other.name}",
        sdfunc=scipy.signal.coherence,
        **kwargs,
    )
    out.override_unit("coherence")
    return out
